﻿CompilerIf Defined(INCLUDE_SPLITTER, #PB_Constant)=0
#INCLUDE_SPLITTER=1
;/////////////////////////////////////////////////////////////////////////////////
;***Splitter*** - Version 1.0.
;Cross platform splitter class.
;
;© Big Eared Software 2017.
;==========================
;   Created with Purebasic 5.61 LTS for Windows.
;
;   Platforms:  All.
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;-General notes.
;
;   i)    This is a crude and unsophisticated splitter implementation (it is not really a proper control!) to try and get around
;         issues with splitters on Ubuntu.
;
;         This splitter does not consist of a parent control in which sit two controls etc. Instead it just handles a canvas based slider control
;         and passes messages back to the host app.
;
;   ii)   It is up to the host app to add margins around the slider etc.
;/////////////////////////////////////////////////////////////////////////////////

EnableExplicit

;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS.

  ;Styles.
    Enumeration
      #Splitter_VERTICAL                    = 1     ;Default is a horizontal splitter.    
    EndEnumeration
    
  ;Event messages.
    Enumeration
      #Splitter_RequestSliderPosChange      ;An instruction to the event proc to validate the new position.
                                            ;Return #True to allow the new position (or altered position).
                                            ;The event handler should hide any gadgets which would be resized to a zero width/height, but do not otherwise resize
                                            ;the gadgets at this point.
                                            ;*x\i = requested X, *y\i = requested Y. 
      #Splitter_SliderPosChanged            ;A notification that the slider has been moved (or the gadget resized) and so all virtual child gadgets should be resized etc.
    EndEnumeration
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-INTERFACES.

  ;The main Splitter class. A pointer to such an object is returned from the NewSplitter() function.
    Interface Splitter
      Destroy()
      GetHeight.i()
      GetWidth.i()
      GetX.i()
      GetY.i()
      Hide(state)
      Resize(x, y, width, height)
    EndInterface
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-PROTOTYPES

  Prototype.i Splitter_ProtoEventHandler(this.Splitter, uMsg, *x.INTEGER, *y.INTEGER)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-STRUCTURES.

  ;Splitter class template.
    Structure _Splitter
      *vTable
      flags.i
      canvasID.i
      ;Event handlers.
        eventProc.Splitter_ProtoEventHandler
    EndStructure
    
   ;Used to hold some global variables, mostly concerned with dragging etc.
    Structure _Splitter_Globals
      blnMouseDown.i
      delta.i
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBALS
  Global gSplitter._Splitter_Globals
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES
    Declare.i NewSplitter(x, y, width, height, eventProc, flags=0)
    Declare Splitter_Function_Paint(*this._Splitter)
    Declare Splitter_SliderEvents()
;/////////////////////////////////////////////////////////////////////////////////


;-CLASS CREATION FUNCTION
;========================

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i NewSplitter(x, y, width, height, eventProc, flags=0)
  Protected *this._Splitter
  If eventProc
    ;Allocate enough memory for our class template.
      *this = AllocateStructure(_Splitter)
    If *this
      *this\vTable = ?Splitter_VT
      *this\eventProc = eventProc
      *this\flags = flags
      *this\canvasID = CanvasGadget(#PB_Any, x, y, width, height)
      If *this\canvasID
        If flags & #Splitter_VERTICAL
          SetGadgetAttribute(*this\canvasID, #PB_Canvas_Cursor, #PB_Cursor_LeftRight)
        Else
          SetGadgetAttribute(*this\canvasID, #PB_Canvas_Cursor, #PB_Cursor_UpDown)
        EndIf
        SetGadgetData(*this\canvasID, *this)
        Splitter_Function_Paint(*this)
        ;Bind events to the slider.
          BindGadgetEvent(*this\canvasID, @Splitter_SliderEvents(), #PB_All)
        ;Instruct the host to resize all relevant gadgets.
        *this\eventProc(*this, #Splitter_SliderPosChanged, @x, @y)
      Else
        FreeStructure(*this)
        *this = 0
      EndIf
    EndIf
  EndIf
  ProcedureReturn *this
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-FUNCTIONS
;==========

;/////////////////////////////////////////////////////////////////////////////////
Procedure Splitter_Function_Paint(*this._Splitter)
  Protected width, height, t1
  If StartDrawing(CanvasOutput(*this\canvasID))
    width = OutputWidth()
    height = OutputHeight()
    Box(0, 0, width, height, $F0F0F0)
    DrawingMode(#PB_2DDrawing_Outlined)
    Box(0, 0, width, height, 0)
    If *this\flags & #Splitter_VERTICAL
      t1 = (height - 30)>>1
      Line(0, t1, width, 1, 0)
      Line(0, t1+10, width, 1, 0)
      Line(0, t1+20, width, 1, 0)
      Line(0, t1+30, width, 1, 0)
    Else ;Horizontal.
      t1 = (width - 30)>>1
      Line(t1, 0, 1, height, 0)
      Line(t1+10, 0, 1, height, 0)
      Line(t1+20, 0, 1, height, 0)
      Line(t1+30, 0, 1, height, 0)
    EndIf
    StopDrawing()
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-EVENT PROCS
;============

;/////////////////////////////////////////////////////////////////////////////////
Procedure Splitter_SliderEvents()
  Protected *this._Splitter, newX, newY
  *this = GetGadgetData(EventGadget())
  If *this
    Select EventType()
      Case #PB_EventType_LeftButtonDown
        If *this\flags & #Splitter_VERTICAL
          gSplitter\blnMouseDown = #True
          gSplitter\delta = GetGadgetAttribute(*this\canvasID, #PB_Canvas_MouseX)
        Else ;Horizontal.
          gSplitter\blnMouseDown = #True
          gSplitter\delta = GetGadgetAttribute(*this\canvasID, #PB_Canvas_MouseY)
        EndIf
      Case #PB_EventType_MouseMove
        If gSplitter\blnMouseDown 
          If *this\flags & #Splitter_VERTICAL
            ;Recalculate X and seek position to move the slider.
              newX = GadgetX(*this\canvasID) + GetGadgetAttribute(*this\canvasID, #PB_Canvas_MouseX) - gSplitter\delta
              newY = GadgetY(*this\canvasID)
              If *this\eventProc(*this, #Splitter_RequestSliderPosChange, @newX, @newY)
                ResizeGadget(*this\canvasID, newX, newY, #PB_Ignore, #PB_Ignore)
                ;Instruct the host to resize all relevant gadgets.
                *this\eventProc(*this, #Splitter_SliderPosChanged, @newX, @newY)
              EndIf
          Else ;Horizontal.
            ;Recalculate Y and seek position to move the slider.
              newX = GadgetX(*this\canvasID)
              newY = GadgetY(*this\canvasID) + GetGadgetAttribute(*this\canvasID, #PB_Canvas_MouseY) - gSplitter\delta
              If *this\eventProc(*this, #Splitter_RequestSliderPosChange, @newX, @newY)
                ResizeGadget(*this\canvasID, newX, newY, #PB_Ignore, #PB_Ignore)
                ;Instruct the host to resize all relevant gadgets.
                *this\eventProc(*this, #Splitter_SliderPosChanged, @newX, @newY)
              EndIf
          EndIf
        EndIf
      Case #PB_EventType_LeftButtonUp
        gSplitter\blnMouseDown = #False
    EndSelect
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-CLASS METHODS
;==============

;/////////////////////////////////////////////////////////////////////////////////
Procedure Splitter_Method_Destroy(*this._Splitter)
  UnbindGadgetEvent(*this\canvasID, @Splitter_SliderEvents(), #PB_All)
  FreeGadget(*this\canvasID)
  FreeStructure(*this)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i Splitter_Method_GetHeight(*this._Splitter)
  ProcedureReturn GadgetHeight(*this\canvasID)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i Splitter_Method_GetWidth(*this._Splitter)
  ProcedureReturn GadgetWidth(*this\canvasID)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i Splitter_Method_GetX(*this._Splitter)
  ProcedureReturn GadgetX(*this\canvasID)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i Splitter_Method_GetY(*this._Splitter)
  ProcedureReturn GadgetY(*this\canvasID)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i Splitter_Method_Hide(*this._Splitter, state)
  HideGadget(*this\canvasID, state)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i Splitter_Method_Resize(*this._Splitter, x, y, width, height)
  If x = #PB_Ignore
    x = GadgetX(*this\canvasID)
  EndIf
  If y = #PB_Ignore
    y = GadgetY(*this\canvasID)
  EndIf
  *this\eventProc(*this, #Splitter_RequestSliderPosChange, @x, @y)
  ResizeGadget(*this\canvasID, x, y, width, height)
  ;Instruct the host to resize all relevant gadgets.
    *this\eventProc(*this, #Splitter_SliderPosChanged, @x, @y)
  Splitter_Function_Paint(*this)
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-VIRTUAL TABLES
;===============

;/////////////////////////////////////////////////////////////////////////////////
  DataSection
    Splitter_VT:
      Data.i @Splitter_Method_Destroy()
      Data.i @Splitter_Method_GetHeight()
      Data.i @Splitter_Method_GetWidth()
      Data.i @Splitter_Method_GetX()
      Data.i @Splitter_Method_GetY()
      Data.i @Splitter_Method_Hide()
      Data.i @Splitter_Method_Resize()
  EndDataSection
;/////////////////////////////////////////////////////////////////////////////////

DisableExplicit

CompilerEndIf
; IDE Options = PureBasic 5.61 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 30
; FirstLine = 30
; Folding = ---
; EnableThread
; Executable = timer.dll
; EnableUnicode