﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Script class.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;/////////////////////////////////////////////////////////////////////////////////



;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES

  Declare.i minScriptCodeClass_HELPER_GetFunctionFromName(name$, *currentModule=0)
  Declare.u minScriptCodeClass_HELPER_String2VT(*this, name$)
  Declare.s minScriptCodeClass_DISASSEMBLY_DisassembleObjectCodeInstruction(*this, opCode, operand.q, *compiler)
  Declare.i minScriptCodeClass_Reset(*this, flags=0)
  Declare.i minScriptCodeClass_DeleteModule(*this, moduleName$)
  Declare.i minScriptCodeClass_ExecuteCodeSnippet(*this, token, blnEnableDebugging=#False)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS

  #minScript_TEMPORARYCODESNIPPETCOMMENT$ = "Executable String"

;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-STRUCTURES

  ;The following is used for the local variables symbol tables and is held in a map within each function's data structure.
  ;Each map uses the variable name (lowercase) as the key.
    Structure st_minScript_LocalVariablesSymbol
      StructureUnion
        EBPoffset.i           ;For use with local variables. Offset from the EBP register.
        isPrivateProperty.i   ;For use with class properties.
      EndStructureUnion
      *staticModuleVar.minScriptVariant   ;Non zero only if we are dealing with a static variable.
      vt.u                ;Contains the #minScript_... variable type.
    EndStructure

  ;The following holds information on individual code functions. These include 'snippets' of code which the host app can create
  ;and compile for execution at any time. 
  ;These structures will be held in a map with a suitable lower case key.
    Structure st_minScriptFunction
      name$
      blnIsPrivate.i            ;Private functions can only be accessed from their parent module.
      *module.st_minScriptModule
      objectCodeBufferPtr.i     ;A pointer to compiled object code. MemorySize() can be used to get the size of the object code.
      ;Local variables.  We place these here so they are persistent allowing any debuggers access etc.
        numLocalVariables.i   ;Includes all optional function parameters, but excludes static variables.
        Map localVariablesSymbols.st_minScript_LocalVariablesSymbol() ;MapSize() gives the actual number of declared variables etc.
      ;String table (holds names of script functions being called as well). The strings are addressed through a 1-based index.
      ;Index 0 is reserved for internal use (typically code snippet comments).
        numStrings.i
        stringTable.s[#minScript_STRINGTABLESIZE+1]
    EndStructure

  ;The following holds information on individual classes. 
  ;These structures will be held in a map with a suitable lower case key.
    Structure st_minScriptClass
      name$
      *code.st_minScriptCodeClass
      *module.st_minScriptModule
      classVT.u           ;Internally generated. Takes the form : #minScript_CLASS + a 12-bit integer. Not unique between code objects.
      numProperties.i
      Map propertyInfo.st_minScript_LocalVariablesSymbol() ;We use this structure for convenience.
      ;The following fields are used to prevent a class constructor/destructor being called in an infinite loop should, for example, a constructor instantiate a new class object etc.
      ;If the threadsafe switch is set then they will hold threadStorage indexes, otherwise straight boolean values.
        tsiConstructorActive.ThreadStorageIndex
        tsiDestructorActive.ThreadStorageIndex
    EndStructure

  ;The following holds information on individual modules. 
  ;These structures will be held in a map with a suitable lower case key.
    Structure st_minScriptModule
      name$
      *code.st_minScriptCodeClass
      ;Module constants.
        Map moduleConstants.st_minScriptGlobalConstant()
      ;Module variables. These are effectively script variables, but defined at the module level.
      ;The following map is indexed on LCase(varName$) and contains the individual variables.
        Map moduleVariables.minScriptVariant() 
      ;Functions.
      Map functions.st_minScriptFunction()  ;Key = LCase(functionName$/token)
    EndStructure

  ;The following is the class template for the minScriptCode class.
    Structure st_minScriptCodeClass
      *vTable
      ;Internal objects.
        *compiler.minScriptCodeCompiler
        *execution.minScriptCodeExecution
        *EAX.minScriptVariant     ;Allows quick access to the EAX register housed in the execution object.
      ;Script variables. ;The following map is indexed on LCase(varName$).
        Map scriptVariables.minScriptVariant()
      ;Modules.
        Map modules.st_minScriptModule()    ;Key = LCase(moduleName$)
        *snippetModule.st_minScriptModule   ;For convenience.
      ;Classes.
        Map classes.st_minScriptClass()
        Map classVTs.u()                    ;Used to keep a consistent (and persistent) record of class VTs so that even if a class is deleted,
                                            ;we can then reuse the VT if a class is then reinstated.
      ;Snippets.
        currentSnippetToken.i               ;Used to create the identifying 'token' and, indirectly, unique map keys for code snippets as described above.
        ;Current function being compiled/executed. Needs to be preserved on a stack etc.
          *currentScriptFunction.st_minScriptFunction
      ;Compiler directives.
        blnEnableExplicit.i ;Always #TRUE for the constantExpressionObject. Only applies to local (non-static) variables.
      numCodeExecutions.i
      ;Debugging.
        blnEnableDebugging.i
        debugger.proto_minScript_debuggerFunction
      ;Errors.
        lastErrorCode.i
        lastErrorLineNumber.i           ;Always refers to a source line even when executing compiled code.
        lastCompilationErrorLine$
        *lastExecutionErrorFunction.st_minScriptFunction
    EndStructure

;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-INCLUDES

  IncludeFile "CLASS_Code_Compiler.pbi"
  IncludeFile "CLASS_Code_Execution.pbi"
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-Create an instance of the minScriptCode class.

;Returns a newly instantiated object or zero if an error (out of memory).
;Flags unused at this time.
Procedure.i minScript_NewCodeObject(flags=0)
  Protected *this.st_minScriptCodeClass, ptrGCExtension.proto_minScript_GCExtension
  ;Attempt to allocate memory for a st_minScriptCodeClass class template.
    *this = AllocateStructure(st_minScriptCodeClass)
  If *this
    *this\vTable = ?VTable_minScriptCodeClass
    ;We add a dummy module for use by code snippets.
    *this\snippetModule = AddMapElement(*this\Modules(), #minScript_SNIPPETMODULE$)
    If *this\snippetModule
      *this\snippetModule\name$ = #minScript_SNIPPETMODULE$
      *this\snippetModule\code = *this
      ;Create the internal objects used for the majority of internal processes.
      *this\compiler = minScript_NewCodeCompilerObject(*this)
      If *this\compiler
        *this\execution = minScript_NewCodeExecutionObject(*this)
        If *this\execution
          *this\blnEnableExplicit = #True
          CompilerIf #PB_Compiler_Thread
            LockMutex(minScript_Globals\threadMutex)
          CompilerEndIf
          ;Increase the global numCodeObjects field if we are dealing with a host created code object.
            If minScriptGlobalClassLib
              minScript_Globals\numCodeObjects + 1
            EndIf
          ;Call all GC extensions with the #minScript_GCEXTENSION_CODEOBJECTHASBEENCREATED message.
          ;We preserve the list position to protect against some GC extension taking some action which might involve this list.
            PushListPosition(minScript_Globals\ptrGCExtensions())
              ForEach minScript_Globals\ptrGCExtensions()
                ptrGCExtension = minScript_Globals\ptrGCExtensions()
                ptrGCExtension(*this, #minScript_GCEXTENSION_CODEOBJECTHASBEENCREATED)
              Next
            PopListPosition(minScript_Globals\ptrGCExtensions())
          CompilerIf #PB_Compiler_Thread
            UnlockMutex(minScript_Globals\threadMutex)
          CompilerEndIf
       Else
          *this\compiler\Destroy()
          Goto minScript_NewCodeObject_L1
        EndIf
      Else
        Goto minScript_NewCodeObject_L1
      EndIf
    Else
minScript_NewCodeObject_L1:
      FreeStructure(*this)
      *this = 0
    EndIf
  EndIf
  ProcedureReturn *this
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS (those not intended for public consumption).

;/////////////////////////////////////////////////////////////////////////////////
;Erase a class. Does not remove the given class from the underlying map.
Procedure minScriptCodeClass_HELPER_EraseClass(*this.st_minScriptCodeClass, *class.st_minScriptClass)
  If *class
    ;We need to run through all module variables (for all modules) and clear any of this class type.
      PushMapPosition(*this\modules())
        ForEach *this\modules()
          ForEach *this\modules()\moduleVariables()
            If *this\modules()\moduleVariables()\vt = *class\classVT And *this\modules()\moduleVariables()\refPtr
              *this\modules()\moduleVariables()\refPtr\Clear() ;Return value can be ignored.
              *this\modules()\moduleVariables()\refPtr = 0
            EndIf
          Next
        Next
      PopMapPosition(*this\modules())
    ;Now repeat for script variables.
      ForEach *this\scriptVariables()
        If *this\scriptVariables()\vt = *class\classVT And *this\scriptVariables()\refPtr
          *this\scriptVariables()\refPtr\Clear() ;Return value can be ignored.
          *this\scriptVariables()\refPtr = 0
        EndIf
      Next
    ;Remove threaded fields if appropriate.
      CompilerIf #PB_Compiler_Thread
        DeleteThreadStorage(*class\tsiConstructorActive)
        DeleteThreadStorage(*class\tsiDestructorActive)
      CompilerEndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Erase a module (does not remove the given module from the underlying map).
Procedure minScriptCodeClass_HELPER_EraseModule(*this.st_minScriptCodeClass, *module.st_minScriptModule, blnEraseClasses=#False)
  If *module
    ;Module constants.
    ;Nothing needed as the structures contain no dynamically allocated memory and will be freed when the module map entry is cleared.
    ;Clear module variables.
    ;We do not need to check the return values because this routine cannot be called during code execution.
      ForEach *module\moduleVariables()
        minScript_VarClear(*module\moduleVariables(), #minScript_VARDONOTCLEARVTFIELD)      
      Next
      ClearMap(*module\moduleVariables())
    ;Remove any functions within the module.
      ForEach *module\functions()
        If *module\functions()\objectCodeBufferPtr
          FreeMemory(*module\functions()\objectCodeBufferPtr)
        EndIf
      Next
      ClearMap(*module\functions())
    ;Remove all classes defined within the module. This essentially entails clearing all module variables whose type matches the given class.
      If blnEraseClasses
        ForEach *this\classes()
          If *this\classes()\module = *module
            minScriptCodeClass_HELPER_EraseClass(*this, *this\classes())
            DeleteMapElement(*this\classes())
          EndIf
        Next
      EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following helper attempts to locate the given function, taking into account any PRIVATE declarations.
;It first tries the 'current' module before looping through the modules in map order.
Procedure.i minScriptCodeClass_HELPER_GetFunctionFromName(name$, *currentModule.st_minScriptModule=0)
  Protected *function.st_minScriptFunction, moduleName$, functionName$, *module.st_minScriptModule
  If name$
    ;Is the name of the form 'moduleName:functionName'
      moduleName$ = StringField(name$, 1, ":")
      functionName$ = StringField(name$, 2, ":")
    CompilerIf #PB_Compiler_Thread
      LockMutex(minScript_Globals\threadMutex)
    CompilerEndIf
    If functionName$
      *module = FindMapElement(*currentModule\code\modules(), moduleName$)
      If *module
        *function = FindMapElement(*module\functions(), functionName$)
        ;Do we have a private function?
          If *function And *function\blnIsPrivate And *module <> *currentModule
           *function = 0
          EndIf
      EndIf      
    Else 
      *function = FindMapElement(*currentModule\functions(), name$)
      If *function = 0
        ForEach *currentModule\code\modules()    
          *function = FindMapElement(*currentModule\code\modules()\functions(), name$)
          If *function And *function\blnIsPrivate=#False
            Break
          EndIf    
          *function = 0
        Next
      EndIf
    EndIf
    CompilerIf #PB_Compiler_Thread
      UnlockMutex(minScript_Globals\threadMutex)
    CompilerEndIf
  EndIf
  ProcedureReturn *function
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;////////////////////////////////////////////////////////////////////////
;The following attempts to convert a string VT type to an actual VT.
  Procedure.u minScriptCodeClass_HELPER_String2VT(*this.st_minScriptCodeClass, name$)
  Protected *gcl.st_minScriptCodeClass, VT.u, *ptr.UNICODE, *ptrClass.st_minScriptClass
  Select name$
    Case "integer"
      VT = #minScript_INTEGER
    Case "currency"
      VT = #minScript_CURRENCY
    Case "real"
      VT = #minScript_REAL
    Case "string"
      VT = #minScript_STRING
    Case "minscriptvariant", "variant"
      VT = #minScript_VARIANT
    Default
      ;We either have a user-defined type (such as an array) or a class (which is still a reference type).
      ;First check for a class defined at the code object level.
        If FindMapElement(*this\classes(), name$)
          VT = *this\classes()\classVT
        Else
          CompilerIf #PB_Compiler_Thread
            LockMutex(minScript_Globals\threadMutex)
          CompilerEndIf
          ;Check for a user-defined type.
          *ptr = FindMapElement(minScript_Globals\userDefinedTypes(), name$)
          If *ptr
            VT = *ptr\u
          Else
            ;We search the global class library.
            *gcl = minScriptGlobalClassLib
            *ptrClass = FindMapElement(*gcl\classes(), name$)
            If *ptrClass
              VT = *ptrClass\classVT
            EndIf
          EndIf
          CompilerIf #PB_Compiler_Thread
            UnlockMutex(minScript_Globals\threadMutex)
          CompilerEndIf
        EndIf
  EndSelect
  ProcedureReturn VT
EndProcedure
;////////////////////////////////////////////////////////////////////////


;-DISASSEMBLY FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_DISASSEMBLY_VT2String(operand, *code.st_minScriptCodeClass)
  Protected VT$, *gcl.st_minScriptCodeClass, *ptr.INTEGER, *ptrClass.st_minScriptClass, vt.u
  vt = operand
  If vt = #minScript_VARIANT
    VT$ = "variant"
  Else
    vt&#minScript_VARIANTMASK
    Select vt
      Case #minScript_EMPTY
        VT$ = "empty" 
      Case #minScript_INTEGER
        VT$ = "integer" 
      Case #minScript_CURRENCY
        VT$ = "currency" 
      Case #minScript_REAL
        VT$ = "real" 
      Case #minScript_STRING
        VT$ = "string" 
      Case  #minScript_BYTE
        VT$ = "byte" 
      Case  #minScript_WORD
        VT$ = "word" 
      Case  #minScript_LONG
        VT$ = "long" 
      Case  #minScript_FLOAT
        VT$ = "float" 
      Case  #minScript_REFERENCE
        VT$ = "reference" 
      Default ;Must be either a user-defined type or a class.
        ;First check for a class in the current object.
          ForEach *code\classes()
            If *code\classes()\classVT = vt
              VT$ = *code\classes()\name$
              Break
            EndIf        
          Next
        If VT$ = ""
          CompilerIf #PB_Compiler_Thread
            LockMutex(minScript_Globals\threadMutex)
          CompilerEndIf
          ;Check for a user-defined type.
          ForEach minScript_Globals\userDefinedTypes()
            If minScript_Globals\userDefinedTypes() = vt
              VT$ = MapKey(minScript_Globals\userDefinedTypes())
              Break
            EndIf
          Next
          If VT$ = ""
            ;Check for a class in the global library object.
            *gcl = minScriptGlobalClassLib
            ForEach *gcl\classes()
              If *gcl\classes()\classVT = vt
                VT$ = *gcl\classes()\name$
                Break
              EndIf        
            Next
          EndIf
          CompilerIf #PB_Compiler_Thread
            UnlockMutex(minScript_Globals\threadMutex)
          CompilerEndIf
        EndIf
    EndSelect
  EndIf
  ProcedureReturn VT$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the name of a code label pointed to by the operand parameter.
Procedure.s minScriptCodeClass_DISASSEMBLY_GetCodeLabel(operand, *compiler.st_minScriptCodeCompilerClass)
  Protected label$
  ;Set a default value.
    label$ = Str(operand)
  ;Iterate through the codelabels map in an attempt to locate the relevant label.
  If *compiler
    ForEach *compiler\codeLabels()
      If operand = @*compiler\codeLabels()
        label$ = MapKey(*compiler\codeLabels())
      EndIf
    Next
  EndIf
  ProcedureReturn label$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the name of a variable relevant to the given opCode.
;*ptr will either be a pointer to a st_minScriptCodeClass structure or a st_minScriptFunction structure.
Procedure.s minScriptCodeClass_DISASSEMBLY_GetNameOfVariable(opCode, operand, *code.st_minScriptCodeClass)
  Protected name$
  Select opCode
    Case #MS_OP_POPLOCALVAR, #MS_OP_PUSHLOCALVAR, #MS_OP_PUSHADDRLOCALVAR
      ;Set a default value.
        name$ = Str(operand)
      ForEach *code\currentScriptFunction\localVariablesSymbols()
        If *code\currentScriptFunction\localVariablesSymbols()\staticModuleVar = 0 And *code\currentScriptFunction\localVariablesSymbols()\EBPoffset = operand
          name$ = MapKey(*code\currentScriptFunction\localVariablesSymbols())
          Break      
        EndIf
      Next
    Case #MS_OP_POPSCRIPTVAR, #MS_OP_PUSHSCRIPTVAR, #MS_OP_PUSHADDRSCRIPTVAR
      ;First try proper script variables.
        ForEach *code\scriptVariables()
          If *code\scriptVariables() = operand
            name$ = MapKey(*code\scriptVariables())
            Break      
          EndIf
        Next
      If name$ = ""
        ;Try a module variable.
        If *code\Modules()
          ForEach *code\Modules()\moduleVariables()
            If *code\Modules()\moduleVariables() = operand
              name$ = MapKey(*code\Modules()\moduleVariables())
              Break      
            EndIf
          Next
        EndIf
        If name$=""
          ;Set a default value.
            name$ = Str(operand)
        EndIf      
      EndIf
  EndSelect
  ProcedureReturn name$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the name of a poKeyword.
Procedure.s minScriptCodeClass_DISASSEMBLY_GetNameOfpoKeyword(operand)
  Protected text$
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ForEach minScript_Globals\minScriptpoKeywords()
    If minScript_Globals\minScriptpoKeywords()\address = operand
      text$ = minScript_Globals\minScriptpoKeywords()\friendlyName$
      Break
    EndIf
  Next
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn text$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following returns the name of a variable relevant to the given opCode.
Procedure.s minScriptCodeClass_DISASSEMBLY_GetStringTableEntry(operand, *code.st_minScriptCodeClass)
  Protected text$
  text$ = Chr(34) + *code\currentScriptFunction\stringTable[operand] + Chr(34)
  ProcedureReturn text$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_DISASSEMBLY_DisassembleObjectCodeInstruction(*this.st_minScriptCodeClass, opCode, operand.q, *compiler.st_minScriptCodeCompilerClass)
  Protected asm$, *ptrDouble.DOUBLE, fnID, numParams, t1
  Select opCode
    Case #MS_OP_DEBUG_NOP
      asm$ = "!NOP (line number:" + Str(operand) + ")" 
    Case #MS_OP_ADD
      asm$ = "!ADD"
    Case #MS_OP_AND
      asm$ = "!AND"
    Case #MS_OP_BITWISEAND
      asm$ = "!BITWISEAND"
    Case #MS_OP_BITWISENOT
      asm$ = "!BITWISENOT"
    Case #MS_OP_BITWISEOR
      asm$ = "!BITWISEOR"
    Case #MS_OP_BITWISESHIFTLEFT
      asm$ = "!BITWISESHIFTLEFT"
    Case #MS_OP_BITWISESHIFTRIGHT
      asm$ = "!BITWISESHIFTRIGHT"
    Case #MS_OP_BITWISEXOR
      asm$ = "!BITWISEXOR"
    Case #MS_OP_CALLFUNCTION
      asm$ = "!CALLFUNCTION " + minScriptCodeClass_DISASSEMBLY_GetStringTableEntry(operand, *this)
    Case #MS_OP_CALLRTFUNCTION
      fnID = operand & $FFFFFFFF
      numParams = operand >> 32
      asm$ = "!CALLRTFUNCTION " + minScript_GetRuntimeFunctionFromID(fnID) + " (" + Str(numParams) + " parameters passed)"
    Case #MS_OP_CAST
      asm$ = "!CAST " + minScriptCodeClass_DISASSEMBLY_VT2String(operand, *this)
    Case #MS_OP_DEBUG
      asm$ = "!DEBUG"
    Case #MS_OP_DIV
      asm$ = "!DIV"
    Case #MS_OP_DUP
      asm$ = "!DUP"
    Case #MS_OP_EBPTOESP
      asm$ = "!EBPTOESP"
    Case #MS_OP_EQU
      asm$ = "!EQU"
    Case #MS_OP_ESPTOEBP
      asm$ = "!ESPTOEBP"
    Case #MS_OP_GT
      asm$ = "!GT"
    Case #MS_OP_GTE
      asm$ = "!GTE"
    Case #MS_OP_METHOD_INVOKE
      fnID = operand & $FFFFFFFF
      numParams = (operand >> 32) & $3FFFFFFF
      asm$ = "!METHODINVOKE " + minScriptCodeClass_DISASSEMBLY_GetStringTableEntry(fnID, *this) + " (" + Str(numParams) + " parameters passed, including *this)"
    Case #MS_OP_METHOD_GET
      fnID = operand & $FFFFFFFF
      numParams = (operand >> 32) & $3FFFFFFF
      asm$ = "!METHODGET " + minScriptCodeClass_DISASSEMBLY_GetStringTableEntry(fnID, *this) + " (" + Str(numParams) + " parameters passed, including *this)"
    Case #MS_OP_METHOD_SET
      fnID = operand & $FFFFFFFF
      numParams = (operand >> 32) & $3FFFFFFF
      asm$ = "!METHODSET " + minScriptCodeClass_DISASSEMBLY_GetStringTableEntry(fnID, *this) + " (" + Str(numParams) + " parameters passed, including *this)"
    Case #MS_OP_METHOD_GETOBJECT
      fnID = operand & $FFFFFFFF
      numParams = (operand >> 32) & $3FFFFFFF
      asm$ = "!METHODGETOBJECT " + minScriptCodeClass_DISASSEMBLY_GetStringTableEntry(fnID, *this) + " (" + Str(numParams) + " parameters passed, including *this)"
    Case #MS_OP_JMP
      asm$ = "!JMP " + minScriptCodeClass_DISASSEMBLY_GetCodeLabel(operand, *compiler)
    Case #MS_OP_JMPNZ
      asm$ = "!JMPNZ " + minScriptCodeClass_DISASSEMBLY_GetCodeLabel(operand, *compiler)
    Case #MS_OP_JMPZ
      asm$ = "!JMPZ " + minScriptCodeClass_DISASSEMBLY_GetCodeLabel(operand, *compiler)
    Case #MS_OP_LOCALPARAM
      asm$ = "!LOCALPARAM " + minScriptCodeClass_DISASSEMBLY_VT2String(operand, *this)
    Case #MS_OP_LOCALVAR
      asm$ = "!LOCALVAR " + minScriptCodeClass_DISASSEMBLY_VT2String(operand, *this)
    Case #MS_OP_LT
      asm$ = "!LT"
    Case #MS_OP_LTE
      asm$ = "!LTE"
    Case #MS_OP_MUL
      asm$ = "!MUL"
    Case #MS_OP_NEG
      asm$ = "!NEG"
    Case #MS_OP_NEQ
      asm$ = "!NEQ"
    Case #MS_OP_OR
      asm$ = "!OR"
    Case #MS_OP_POKEYWORD0
      asm$ = "!POKEYWORD0 " + minScriptCodeClass_DISASSEMBLY_GetNameOfpoKeyword(operand)
    Case #MS_OP_POKEYWORD1
      asm$ = "!POKEYWORD1 " + minScriptCodeClass_DISASSEMBLY_GetNameOfpoKeyword(operand)
    Case #MS_OP_POP
      asm$ = "!POP " + Str(operand)
    Case #MS_OP_POPEAX
      asm$ = "!POPEAX"
    Case #MS_OP_POPEBP
      asm$ = "!POPEBP"
    Case #MS_OP_POPLOCALVAR
      asm$ = "!POPLOCALVAR " + minScriptCodeClass_DISASSEMBLY_GetNameOfVariable(opCode, operand, *this)
    Case #MS_OP_POPSCRIPTVAR
      asm$ = "!POPSCRIPTVAR " + minScriptCodeClass_DISASSEMBLY_GetNameOfVariable(opCode, operand, *this)
    Case #MS_OP_PUSHADDRLOCALVAR
      asm$ = "!PUSHADDRLOCALVAR " + minScriptCodeClass_DISASSEMBLY_GetNameOfVariable(opCode, operand, *this)
    Case #MS_OP_PUSHADDRSCRIPTVAR
      asm$ = "!PUSHADDRSCRIPTVAR " + minScriptCodeClass_DISASSEMBLY_GetNameOfVariable(opCode, operand, *this)
    Case #MS_OP_PUSHCUR
      asm$ = "!PUSHCUR " + StrC(operand)
    Case #MS_OP_PUSHEAX
      asm$ = "!PUSHEAX"
    Case #MS_OP_PUSHEBP
      asm$ = "!PUSHEBP"
    Case #MS_OP_PUSHINT
      asm$ = "!PUSHINT " + Str(operand)
    Case #MS_OP_PUSHLOCALVAR
      asm$ = "!PUSHLOCALVAR " + minScriptCodeClass_DISASSEMBLY_GetNameOfVariable(opCode, operand, *this)
    Case #MS_OP_PUSHNULL
      asm$ = "!PUSHNULL"
    Case #MS_OP_PUSHREAL
      ;Operand is stored in quad form so we need to convert it for disassembly. We use a pointer.
      *ptrDouble = @operand
      asm$ = "!PUSHREAL " + StrD(*ptrDouble\d)
    Case #MS_OP_PUSHSCRIPTVAR
      asm$ = "!PUSHSCRIPTVAR " + minScriptCodeClass_DISASSEMBLY_GetNameOfVariable(opCode, operand, *this)
    Case #MS_OP_PUSHSTR
      asm$ = "!PUSHSTR " + minScriptCodeClass_DISASSEMBLY_GetStringTableEntry(operand, *this)
    Case #MS_OP_PUSHADDRSTR
      asm$ = "!PUSHADDRSTR " + minScriptCodeClass_DISASSEMBLY_GetStringTableEntry(operand, *this)
    Case #MS_OP_RET
      asm$ = "!RET"
    Case #MS_OP_SUB
      asm$ = "!SUB"
    Case #MS_OP_XOR
      asm$ = "!XOR"
  EndSelect
  ProcedureReturn asm$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-CLASS METHODS (PUBLIC).

;/////////////////////////////////////////////////////////////////////////////////
Procedure minScriptCodeClass_Destroy(*this.st_minScriptCodeClass)
  Protected ptrGCExtension.proto_minScript_GCExtension
  ;Do not allow the user to destroy the class library object. For this case only allow a reset.
  If *this\numCodeExecutions = 0
    If *this = minScriptGlobalClassLib
      minScriptCodeClass_Reset(*this, #minScript_RESET_REMOVEALL)
    Else
      ;Erase modules.
        ForEach *this\Modules()
          minScriptCodeClass_HELPER_EraseModule(*this, *this\Modules()) ;Return value can be ignored as we are destroying all objects anyhow.
        Next
      ;Clear script variables.
        ForEach *this\scriptVariables()
          minScript_VarClear(*this\scriptVariables())
        Next
      ;Clear classes if appropriate.
        CompilerIf #PB_Compiler_Thread
          ForEach *this\classes()
            DeleteThreadStorage(*this\classes()\tsiConstructorActive)
            DeleteThreadStorage(*this\classes()\tsiDestructorActive)
          Next
        CompilerEndIf
      ;Destroy internal objects.
        *this\execution\Destroy()
        *this\compiler\Destroy()
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScript_Globals\threadMutex)
      CompilerEndIf
      ;Decrease the global numCodeObjects field.
        minScript_Globals\numCodeObjects - 1
      ;Call all GC extensions with the #minScript_GCEXTENSION_CODEOBJECTHASBEENDELETED message.
      ;We preserve the list position to protect against some GC extension taking some action which might involve this list.
        PushListPosition(minScript_Globals\ptrGCExtensions())
          ForEach minScript_Globals\ptrGCExtensions()
            ptrGCExtension = minScript_Globals\ptrGCExtensions()
            ptrGCExtension(*this, #minScript_GCEXTENSION_CODEOBJECTHASBEENDELETED)
          Next
        PopListPosition(minScript_Globals\ptrGCExtensions())
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScript_Globals\threadMutex)
      CompilerEndIf
      FreeStructure(*this)
    EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_Reset(*this.st_minScriptCodeClass, flags=0)
  Protected result = #minScript_OKAY, ptrGCExtension.proto_minScript_GCExtension
  If *this\numCodeExecutions = 0 And (*this <> minScriptGlobalClassLib Or minScript_Globals\numCodeObjects = 0) ;Can get away without locking the global mutex.
    If flags & #minScript_RESET_REMOVEALLMODULESANDSNIPPETSANDCLASSES
      ;Remove all modules (does not remove the snippet module map entry).
        ForEach *this\Modules()
          minScriptCodeClass_HELPER_EraseModule(*this, *this\Modules())
          If *this\Modules() <> *this\snippetModule
            DeleteMapElement(*this\Modules())
          EndIf
        Next
      *this\currentSnippetToken = 0
      ;Remove classes.
      ;First remove the threaded fields if appropriate.
        CompilerIf #PB_Compiler_Thread
          ForEach *this\classes()
            DeleteThreadStorage(*this\classes()\tsiConstructorActive)
            DeleteThreadStorage(*this\classes()\tsiDestructorActive)
          Next
        CompilerEndIf
        ClearMap(*this\classes())
        ClearMap(*this\classVTs())
    EndIf
    If flags & #minScript_RESET_REMOVEALLSCRIPTVARIABLES
      ForEach *this\scriptVariables()
        minScript_VarClear(*this\scriptVariables())
      Next
      ClearMap(*this\scriptVariables())
    EndIf
    *this\execution\Reset(flags)
    *this\compiler\Reset(flags)
    ;In the case that the object has been completely reset, we call all GC extensions with the #minScript_GCEXTENSION_CODEOBJECTHASBEENRESET message.
    ;We preserve the list position to protect against some GC extension taking some action which might involve this list.
      If flags = #minScript_RESET_REMOVEALL
        CompilerIf #PB_Compiler_Thread
          LockMutex(minScript_Globals\threadMutex)
        CompilerEndIf
        PushListPosition(minScript_Globals\ptrGCExtensions())
          ForEach minScript_Globals\ptrGCExtensions()
            ptrGCExtension = minScript_Globals\ptrGCExtensions()
            ptrGCExtension(*this, #minScript_GCEXTENSION_CODEOBJECTHASBEENRESET)
          Next
        PopListPosition(minScript_Globals\ptrGCExtensions())
        CompilerIf #PB_Compiler_Thread
          UnlockMutex(minScript_Globals\threadMutex)
        CompilerEndIf
      EndIf
  Else
    result = #minScript_ACCESSDENIED
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_AddScriptVariable(*this.st_minScriptCodeClass, name$, VT$)
  Protected result = #minScript_OKAY, *ptr.CHARACTER, t1$, vt.u, *ptrVariant.minScriptVariant
  If *this <> minScriptGlobalClassLib
    If name$ 
      If VT$
        ;Check the variable name and prepend it with a $ character if required.
        *ptr = @name$
        If *ptr\c = '$'
          t1$ = LCase(name$)
          *ptr + SizeOf(CHARACTER)
        Else
          t1$ = "$" + LCase(name$)
        EndIf
        result = minScript_SymbolsIsSymbolNameValid(*ptr)
        If result = #minScript_OKAY
          ;Locate the VT type.
            vt = minScriptCodeClass_HELPER_String2VT(*this, LCase(VT$))
          If vt
            ;Do we already have a variable defined.
              *ptrVariant = FindMapElement(*this\scriptVariables(), t1$)
            If *ptrVariant
              result = minScript_VarClear(*ptrVariant)
              If result = #minScript_OKAY
                *ptrVariant\vt =  vt
              EndIf
            Else
              ;Add a new variable to the script functions map.
                *ptrVariant = AddMapElement(*this\scriptVariables(), t1$)
              If *ptrVariant
                *ptrVariant\vt =  vt
              Else
                result = #minScript_OUTOFMEMORY           
              EndIf
            EndIf
          Else
            result = #minScript_INVALIDDATATYPE   
          EndIf
        EndIf
      Else
        result = #minScript_INVALIDDATATYPE   
      EndIf
    Else
      result = #minScript_INVALIDSYMBOLNAME
    EndIf
  Else
    result = #minScript_ACCESSDENIED  
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_DeleteScriptVariable(*this.st_minScriptCodeClass, name$)
  Protected result = #minScript_OKAY, *ptrVariant.minScriptVariant
  If *this\numCodeExecutions = 0
    If name$
      name$ = LCase(name$)
      If Left(name$,1) <> "$"
        name$ = "$" + name$    
      EndIf
      ;We need to clear the variable.
      *ptrVariant = FindMapElement(*this\scriptVariables(), name$)
      If *ptrVariant
        result = minScript_VarClear(*ptrVariant)
        DeleteMapElement(*this\scriptVariables())
      EndIf
    Else
      result = #minScript_INVALIDSYMBOLNAME
    EndIf
  Else
    result = #minScript_ACCESSDENIED
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_GetListOfScriptVariables(*this.st_minScriptCodeClass)
  Protected result$
  ForEach *this\scriptVariables()
    result$ + MapKey(*this\scriptVariables()) + #TAB$  
  Next
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_GetScriptVariablePtr(*this.st_minScriptCodeClass, name$)
  Protected *ptrVariant.minScriptVariant
  If name$
    name$ = LCase(name$)
    If Left(name$,1) <> "$"
      name$ = "$" + name$    
    EndIf
    *ptrVariant = FindMapElement(*this\scriptVariables(), name$)
  EndIf
  ProcedureReturn *ptrVariant
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_GetListOfClasses(*this.st_minScriptCodeClass)
  Protected result$
  ForEach *this\classes()
    result$ + *this\classes()\name$ + #TAB$  
  Next
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_AddModule(*this.st_minScriptCodeClass, moduleName$, *ptrCode.CHARACTER)
  Protected result = #minScript_OKAY, name$, *module.st_minScriptModule
  If *this\numCodeExecutions = 0 And (*this <> minScriptGlobalClassLib Or minScript_Globals\numCodeObjects = 0) ;Can get away without locking the global mutex.
    If moduleName$ And *ptrCode
      name$ = LCase(moduleName$)
      result = minScript_SymbolsIsSymbolNameValid(@name$)
      If result = #minScript_OKAY
        result = minScriptCodeClass_DeleteModule(*this, name$)
        If result = #minScript_OKAY Or result = #minScript_MODULENOTFOUND
          ;Attempt to create a new module.
          *module = AddMapElement(*this\Modules(), name$)
          If *module
            *module\name$ = moduleName$
            *module\code = *this
            If *ptrCode\c
              ;Hand over to the compiler. It will return at the end of the source or when an error is encountered.
              ;Any code outside of a function block will result in an error.
                result = *this\compiler\CompileCode(*ptrCode, *module)
              If result <> #minScript_OKAY
                ;Delete any trace of the new module.
                  minScriptCodeClass_DeleteModule(*this, name$)
              EndIf
            EndIf 
          Else
            result = #minScript_OUTOFMEMORY
          EndIf
        EndIf
      EndIf
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_ACCESSDENIED
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_DeleteModule(*this.st_minScriptCodeClass, moduleName$)
  Protected result = #minScript_OKAY, *module.st_minScriptModule
  If *this\numCodeExecutions = 0 And (*this <> minScriptGlobalClassLib Or minScript_Globals\numCodeObjects = 0) ;Can get away without locking the global mutex.
    If moduleName$
      moduleName$ = LCase(moduleName$)
      *module = FindMapElement(*this\Modules(), moduleName$)
      If *module
        minScriptCodeClass_HELPER_EraseModule(*this, *Module, #True) ;#True to erase all classes defined in the module.
        If *module <> *this\snippetModule
          DeleteMapElement(*this\Modules())        
        EndIf
      Else
        result = #minScript_MODULENOTFOUND
      EndIf
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_ACCESSDENIED
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_GetListOfModules(*this.st_minScriptCodeClass)
  Protected result$
  ForEach *this\Modules()
    If *this\Modules() <> *this\snippetModule
      result$ + *this\Modules()\name$ + #TAB$  
    EndIf
  Next
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_GetListOfModuleFunctions(*this.st_minScriptCodeClass, moduleName$)
  Protected result$, *module.st_minScriptModule
  moduleName$ = LCase(moduleName$)
  If moduleName$ 
    *module = FindMapElement(*this\Modules(), moduleName$)
    If *module
      ForEach *module\functions()
        result$ + *module\functions()\name$ + #TAB$  
      Next
    EndIf
  EndIf
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_AddCodeSnippet(*this.st_minScriptCodeClass, *ptrCode.CHARACTER, *token.INTEGER, comment$="")
  Protected result = #minScript_OKAY, token, name$
  If *this\numCodeExecutions = 0 And *this <> minScriptGlobalClassLib
    If *token
      token = *token\i
      ;If *token references an existing snippet then we must remove the underlying object code.
      If token
        name$ = Str(token) + "_snippet"
        If FindMapElement(*this\snippetModule\functions(), name$)
          If *this\snippetModule\functions()\objectCodeBufferPtr
            FreeMemory(*this\snippetModule\functions()\objectCodeBufferPtr)
            *this\snippetModule\functions()\objectCodeBufferPtr = 0
          EndIf
        EndIf
      Else ;Allocate a new token.
        token = *this\currentSnippetToken + 1
        name$ = Str(token) + "_snippet"
      EndIf
      ;Add a new function to represent the snippet.
        *this\currentScriptFunction = AddMapElement(*this\snippetModule\functions(), name$)
        If *this\currentScriptFunction
          *this\currentScriptFunction\name$ = name$
          *this\currentScriptFunction\module = *this\snippetModule
          *this\currentScriptFunction\stringTable[0] = comment$
          If *ptrCode\c
            ;Hand over to the compiler. It will return at the end of the source or when an error is encountered.
            ;Any code inside a function block will result in an error.
              result = *this\compiler\CompileCode(*ptrCode, *this\snippetModule)
          EndIf
        Else
          result = #minScript_OUTOFMEMORY      
        EndIf
        If result = #minScript_OKAY And *token\i=0
          *this\currentSnippetToken = token
          *token\i = token
        EndIf
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf 
  Else
    result = #minScript_ACCESSDENIED
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_DeleteCodeSnippet(*this.st_minScriptCodeClass, *token.INTEGER)
  Protected result = #minScript_OKAY
  If *this\numCodeExecutions = 0
    If *token
      If *token\i
        If FindMapElement(*this\snippetModule\functions(), Str(*token\i) + "_snippet")
          If *this\snippetModule\functions()\objectCodeBufferPtr
            FreeMemory(*this\snippetModule\functions()\objectCodeBufferPtr)
          EndIf
          ;Delete the map element.
            DeleteMapElement(*this\snippetModule\functions())
          *token\i = 0    
        Else
          result = #minScript_INVALIDTOKEN
        EndIf
      EndIf
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf 
  Else
    result = #minScript_ACCESSDENIED
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_Execute(*this.st_minScriptCodeClass, code$, blnEnableDebugging=#False)
  Protected result = #minScript_OKAY, *tempScriptFunction.st_minScriptFunction, token, t1, tCurrentSnippetToken
  If *this <> minScriptGlobalClassLib
    If code$
      ;We perform a reset if this is an 'initial' execution (as opposed to one embedded within a debugger or run-time function etc.)
        If *this\numCodeExecutions = 0
          minScriptCodeClass_Reset(*this, 0)
        EndIf
      If *this\lastErrorCode = #minScript_OKAY
        ;We create a temporary code snippet.
        ;We don't need to make a permanent alteration to the value of *this\currentSnippetToken and so we make a copy and reinstate later.
            tCurrentSnippetToken = *this\currentSnippetToken
          *this\currentSnippetToken + 1
          *tempScriptFunction = AddMapElement(*this\snippetModule\functions(), Str(*this\currentSnippetToken) + "_snippet")
          If *tempScriptFunction
            *tempScriptFunction\module = *this\snippetModule
            *tempScriptFunction\stringTable[0] = #minScript_TEMPORARYCODESNIPPETCOMMENT$
            ;Hand over to the compiler. It will return at the end of the source or when an error is encountered.
            ;Any code inside a function block will result in an error.
            ;We make a temporary copy of the \currentScriptFunction field in case this method is invoked by a run-time function or user-defined object method
            ;as compilation will reset this field.
              t1 = *this\currentScriptFunction 
              *this\currentScriptFunction = *tempScriptFunction
              result = *this\compiler\CompileCode(@code$, *this\snippetModule)
              *this\currentScriptFunction = t1
              If result = #minScript_OKAY
                ;Execute object code.
                  *this\blnEnableDebugging = blnEnableDebugging
                  result = *this\execution\Execute(*tempScriptFunction)
                ;In the case that an execution error originated from the temporary snippet we have to clear the *this\lastExecutionErrorFunction field since
                ;this temporary function will be deleted.
                  If *this\lastExecutionErrorFunction = *tempScriptFunction
                    *this\lastExecutionErrorFunction = 0
                  EndIf
                ;Delete the temporary function entry.
                  If *tempScriptFunction\objectCodeBufferPtr
                    FreeMemory(*tempScriptFunction\objectCodeBufferPtr)
                  EndIf
                  DeleteMapElement(*this\snippetModule\functions(), Str(*this\currentSnippetToken) + "_snippet")
              EndIf
          Else
            result = #minScript_OUTOFMEMORY      
            *this\lastErrorCode = result
            *this\lastErrorLineNumber = 0
            *this\lastCompilationErrorLine$ = ""
            *this\lastExecutionErrorFunction = 0
          EndIf
         ;Reinstate the current snippet token.
          *this\currentSnippetToken = tCurrentSnippetToken
      EndIf
    EndIf
  Else
    result = #minScript_ACCESSDENIED
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_ExecuteCodeSnippet(*this.st_minScriptCodeClass, token, blnEnableDebugging=#False)
  Protected result = #minScript_OKAY
  ;We perform a reset if this is an 'initial' execution (as opposed to one embedded within a debugger or run-time function etc.)
    If *this\numCodeExecutions = 0
      minScriptCodeClass_Reset(*this, 0)
    EndIf
  If *this\lastErrorCode = #minScript_OKAY
    If token And FindMapElement(*this\snippetModule\functions(), Str(token) + "_snippet")
      ;Execute object code.
        *this\blnEnableDebugging = blnEnableDebugging
        result = *this\execution\Execute(*this\snippetModule\functions())
    Else
      result = #minScript_INVALIDTOKEN
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_GetCodeSnippetComment(*this.st_minScriptCodeClass, token)
  If token And FindMapElement(*this\snippetModule\functions(), Str(token) + "_snippet")
    ProcedureReturn *this\snippetModule\functions()\stringTable[0]
  EndIf  
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_GetReturnVariantPtr(*this.st_minScriptCodeClass)
  ProcedureReturn *this\EAX
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_IsCodeBeingExecuted(*this.st_minScriptCodeClass)
  ProcedureReturn *this\numCodeExecutions
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure minScriptCodeClass_SetDebuggerCallback(*this.st_minScriptCodeClass, address)
  If *this <> minScriptGlobalClassLib
    *this\debugger = address
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_GetLastErrorCode(*this.st_minScriptCodeClass)
  ProcedureReturn *this\lastErrorCode
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_GetLastErrorLineNumber(*this.st_minScriptCodeClass)
  ProcedureReturn *this\lastErrorLineNumber
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_GetLastCompilationErrorLine(*this.st_minScriptCodeClass)
  ProcedureReturn *this\lastCompilationErrorLine$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.s minScriptCodeClass_GetLastExecutionErrorFunctionName(*this.st_minScriptCodeClass)
  Protected name$
  If *this\lastExecutionErrorFunction
    If *this\lastExecutionErrorFunction\module\code = minScriptGlobalClassLib
      name$ = "GCL->"
    EndIf
    name$ + *this\lastExecutionErrorFunction\module\name$ + ":" + *this\lastExecutionErrorFunction\name$
  EndIf
  ProcedureReturn name$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptCodeClass_GetLastExecutionErrorFunction(*this.st_minScriptCodeClass)
  ProcedureReturn *this\lastExecutionErrorFunction
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


DataSection 
  VTable_minScriptCodeClass: 
    Data.i @minScriptCodeClass_Destroy()
    Data.i @minScriptCodeClass_Reset()
    Data.i @minScriptCodeClass_AddScriptVariable()
    Data.i @minScriptCodeClass_DeleteScriptVariable()
    Data.i @minScriptCodeClass_GetListOfScriptVariables()
    Data.i @minScriptCodeClass_GetScriptVariablePtr()
    Data.i @minScriptCodeClass_GetListOfClasses()
    Data.i @minScriptCodeClass_AddModule()
    Data.i @minScriptCodeClass_DeleteModule()
    Data.i @minScriptCodeClass_GetListOfModules()
    Data.i @minScriptCodeClass_GetListOfModuleFunctions()
    Data.i @minScriptCodeClass_AddCodeSnippet()
    Data.i @minScriptCodeClass_DeleteCodeSnippet()
    Data.i @minScriptCodeClass_Execute()
    Data.i @minScriptCodeClass_ExecuteCodeSnippet()
    Data.i @minScriptCodeClass_GetCodeSnippetComment()
    Data.i @minScriptCodeClass_GetReturnVariantPtr()
    Data.i @minScriptCodeClass_IsCodeBeingExecuted()
    Data.i @minScriptCodeClass_SetDebuggerCallback()
    Data.i @minScriptCodeClass_GetLastErrorCode()
    Data.i @minScriptCodeClass_GetLastErrorLineNumber()
    Data.i @minScriptCodeClass_GetLastCompilationErrorLine()
    Data.i @minScriptCodeClass_GetLastExecutionErrorFunctionName()
    Data.i @minScriptCodeClass_GetLastExecutionErrorFunction()
EndDataSection

; IDE Options = PureBasic 5.62 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 163
; FirstLine = 152
; Folding = -------
; Markers = 333
; Executable = d.exe.dll.exe.dll