﻿CompilerIf Defined(INCLUDE_MEMORYFILE, #PB_Constant)=0
#INCLUDE_MEMORYFILE=1
;/////////////////////////////////////////////////////////////////////////////////
;An OOP class for implementing a dynamic memory file intended for binary files.
;For a heavy-weight version of this utility see our qDoc library.

;nxSoftware.com
;Updated 2016.

;Developed with Purebasic 5.42.
;Fully cross-platform.
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;To create an instance of a new memory file class, use something like :
;
;   MyMemoryFileObject.MemoryFile
;   MyMemoryFileObject = NewMemoryFile(initialSize.i, pageSize.l, flags=#MEMORYFILE_VARIABLEBUFFER)
;
;and check the return value is non zero.
;
;NOTE that file writes can (if using a growable buffer) result in the file base address changing. Keep this in mind if using the
;\GetFileMemoryBase() method etc.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-PUBLIC CONSTANTS.

  ;The following creation flags can be combined with the | operator.
    #MEMORYFILE_VARIABLEBUFFER  = 1   ;Default. Allows the underlying buffer to grow and, if required, move in memory.
    #MEMORYFILE_FIXEDBUFFER     = 2   ;Prohibits the underlying buffer from growing (and hence possibly moving).
    #MEMORYFILE_MAKETHREADSAFE  = 4   ;Uses mutex protection to protect the shared resources if multiple threads
                                      ;are being used to access a single memory file etc.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-PUBLIC INTERFACES.

  ;The following interface contains the base class methods.
    Interface MemoryFile
      CheckForErrors.i()                    ;Returns #True if an error has been recorded against any 'write' operation (since the last
                                            ;call to ClearErrorFlag()).
                                            ;Used when checking for errors after a 'bulk' set of write operations etc. as opposed to
                                            ;checking each individual 'write'. A failed 'write' will not cause a crash.
      ClearErrorFlag()
      Destroy.i(blnFreeFileMemory=#True)    ;Set blnFreeFileMemory to #False if the underlying memory is not to be freed.
                                            ;In this case, the return value is a pointer to the underlying memory which can be freed later
                                            ;with FreeMemory() etc. Use the Lof() method first if required.
      Eof.i()                               ;Returns #True if the file pointer is at or past the end of the file.
      FileSeek.i(newOffset.i)               ;Relative to the beginning of the file.
                                            ;The file will be extended and null-padded if the file-pointer is moved beyond the end of the file etc.
                                            ;Returns zero if error.
      GetFileMemoryBase.i()                 ;THIS ADDRESS CAN CHANGE AFTER USING ANY OF THE 'Write...()' METHODS!!!
      GetPageSize.i()                       ;Returns the page size.
      Loc.i()                               ;Returns the current file position as an offset from the beginning of the file.
      Lof.i()
      ;Read methods----------------------------------------------------------------------------------------------------------------------
      ;These all read from the current file position.
        ReadByte.b()
        ReadCharacter.c()
        ReadData.i(memoryBuffer, lengthToRead)  ;Returns the number of bytes actually read.
        ReadDouble.d()
        ReadFloat.f()
        ReadInteger.i()
        ReadLong.l()
        ReadQuad.q()
        ReadString.s(format=0)              ;Omit format to read the string in unicode if the program is compiled in unicode mode or ascii...
                                            ;Otherwise format can take one of the values #PB_Unicode, #PB_UTF8 or #PB_Ascii.
                                            ;Reading finishes when the appropriate null-terminator is encountered (or the end of the file).
        ReadWord.w()
      ;-----------------------------------------------------------------------------------------------------------------------------------
      Reset.i(newInitialSize=-1)            ;Clears the data, resets the file size to the new size, zeros the file pointer,
                                            ;clears the error flag etc.
                                            ;Returns zero if error. In which case the existing memory file is left unchanged.
      SetPageSize(pageSize)                 ;No return value.
      Truncate()                            ;Discards the file from the current file position onwards, thus shortening the file.
      ;Write methods---------------------------------------------------------------------------------------------------------------------
      ;These all write starting at the current file position.
        WriteByte.i(byte.b)                 ;Returns zero if error. The only possible error is if extra memory is required and the subsequent memory re-allocation fails.
        WriteCharacter.i(char.c)            ;Returns zero if error. The only possible error is if extra memory is required and the subsequent memory re-allocation fails.
        WriteData.i(memoryBuffer, length)   ;Returns zero if error.
        WriteDouble.i(double.d)             ;Returns zero if error. The only possible error is if extra memory is required and the subsequent memory re-allocation fails.
        WriteFloat.i(long.f)                ;Returns zero if error. The only possible error is if extra memory is required and the subsequent memory re-allocation fails.
        WriteInteger.i(integer.i)           ;Returns zero if error. The only possible error is if extra memory is required and the subsequent memory re-allocation fails.
        WriteLong.i(long.l)                 ;Returns zero if error. The only possible error is if extra memory is required and the subsequent memory re-allocation fails.
        WriteQuad.i(quad.q)                 ;Returns zero if error. The only possible error is if extra memory is required and the subsequent memory re-allocation fails.
        WriteString.i(text$, format=0)      ;Omit format to write the string in unicode if the program is compiled in unicode mode or ascii...
                                            ;Otherwise format can take one of the values #PB_Unicode, #PB_UTF8 or #PB_Ascii.
                                            ;The appropriate null-terminator is written.
        WriteWord.i(word.w)                 ;Returns zero if error. The only possible error is if extra memory is required and the subsequent memory re-allocation fails.
      ;-----------------------------------------------------------------------------------------------------------------------------------
    EndInterface
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;-MEMORYFILE CREATION FUNCTIONS.

    Declare.i NewMemoryFile(initialSize.i, pageSize.l, flags=#MEMORYFILE_VARIABLEBUFFER)
    Declare.i NewMemoryFileFromMemoryBuffer(*buffer, bufferSize, pageSize.l, flags=#MEMORYFILE_VARIABLEBUFFER)
;/////////////////////////////////////////////////////////////////////////////////


EnableExplicit

;The following structure contains the class template and private properties,
  Structure _membersMemoryFileClass
    *vTable
    flags.i
    fileBase.i
    fileSize.i
    initialSize.i
    initialPageSize.i
    usedFile.i    ;Lof.
    filePointer.i ;Loc.
    mutex.i
    pageSize.l
    blnErrorReported.l  ;#True or #False. Used when checking for errors after a 'bulk' set of write operations etc.
                        ;as opposed to checking each individual 'write'. A failed 'write' will not cause a crash.
  EndStructure 


Declare.i memoryFileClass_AllocateExtraBytes(*this._membersMemoryFileClass, numBytes.i)


;-FUNCTIONS FOR CREATING NEW INSTANCES OF THE MEMORYFILE CLASS.

;/////////////////////////////////////////////////////////////////////////////////
;Returns zero if an error.
Procedure.i NewMemoryFile(initialSize.i, pageSize.l, flags=#MEMORYFILE_VARIABLEBUFFER)
  Protected *object._membersMemoryFileClass, base
  If initialSize>0 And pageSize>0
    ;Allocate memory for the object itself.
      *object = AllocateMemory(SizeOf(_membersMemoryFileClass))
      If *object
        ;Attempt to allocate the initial memory.
          *object\FileBase = AllocateMemory(initialSize)
          If *object\FileBase
            *object\flags = flags
            ;Create a mutex if appropriate.
              If flags & #MEMORYFILE_MAKETHREADSAFE
                *object\mutex = CreateMutex()
                If *object\mutex = 0
                  FreeMemory(*object\FileBase)
                  FreeMemory(*object)
                  ProcedureReturn 0 ;Error.
                EndIf                  
              EndIf
            ;Set the object properties.
              With *object
                \vTable = ?VTable_MemoryFileClass
                \FileSize = initialSize
                \initialSize = initialSize
                \initialPageSize = pageSize
                \usedFile = 0
                \pageSize = pageSize
              EndWith
          Else
            FreeMemory(*object)
            *object = 0
          EndIf
      EndIf
  EndIf
  ProcedureReturn *object
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Returns zero if an error.
Procedure.i NewMemoryFileFromMemoryBuffer(*buffer, bufferSize, pageSize.l, flags=#MEMORYFILE_VARIABLEBUFFER)
  Protected *object._membersMemoryFileClass, base
  If *buffer And bufferSize>0 And pageSize>0
    ;Allocate memory for the object itself.
      *object = AllocateMemory(SizeOf(_membersMemoryFileClass))
    If *object
      *object\FileBase = *buffer
      *object\flags = flags
      ;Create a mutex if appropriate.
        If flags & #MEMORYFILE_MAKETHREADSAFE
          *object\mutex = CreateMutex()
          If *object\mutex = 0
            FreeMemory(*object)
            ProcedureReturn 0 ;Error.
          EndIf                  
        EndIf
      ;Set the object properties.
        With *object
          \vTable = ?VTable_MemoryFileClass
          \FileSize = bufferSize
          \initialSize = bufferSize
          \initialPageSize = pageSize
          \usedFile = bufferSize
          \pageSize = pageSize
        EndWith
      EndIf
  EndIf
  ProcedureReturn *object
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-CLASS METHODS.

;/////////////////////////////////////////////////////////////////////////////////
;The following method returns #True if there has been a 'write' error recorded since the last ClearErrorFlag() call.
Procedure.i memoryFileClass_CheckForErrors(*this._membersMemoryFileClass)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = *this\blnErrorReported
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method simply clears the error status of the underlying memory file.
Procedure memoryFileClass_ClearErrorFlag(*this._membersMemoryFileClass)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  *this\blnErrorReported = #False
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method frees, if requested, all memory associated with the underlying memory File object.
;Returns, in the case that blnFreeFileMemory=#False, the address of the underlying memory File.
Procedure.i memoryFileClass_Destroy(*this._membersMemoryFileClass, blnFreeFileMemory=#True)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If blnFreeFileMemory
    FreeMemory(*this\FileBase)
  Else
    result = *this\FileBase
  EndIf
  ;Now any mutex.
    If *this\mutex
      UnlockMutex(*this\mutex)
      FreeMutex(*this\mutex)
    EndIf
  ;Now the object itself.
    FreeMemory(*this)
    *this = 0
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method returns #True if the file pointer is past the end of the file.
Procedure.i memoryFileClass_Eof(*this._membersMemoryFileClass)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer>=*this\usedFile
    result = #True
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method repositions the read/write position within the underlying memory file.
;Returns non-zero if succesful.
Procedure.i memoryFileClass_FileSeek(*this._membersMemoryFileClass, newOffset.i)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If newOffset >= 0 And newOffset <= *this\usedFile
    *this\filePointer = newOffset
    result = 1
  ElseIf newOffset > *this\usedFile ;Here we may have to enlarge the file.
    result = memoryFileClass_AllocateExtraBytes(*this, newOffset-*this\filePointer)
    If result
      *this\filePointer = newOffset
      *this\usedFile = newOffset
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method returns the underlying memory file address.
Procedure.i memoryFileClass_GetFileMemoryBase(*this._membersMemoryFileClass)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = *this\FileBase
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method returns the page size.
Procedure.i memoryFileClass_GetPageSize(*this._membersMemoryFileClass)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = *this\pageSize
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method returns the current file pointer as a memory address.
Procedure.i memoryFileClass_Loc(*this._membersMemoryFileClass)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = *this\filePointer
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method returns the length (bytes) of the used section of the memory file.
Procedure.i memoryFileClass_Lof(*this._membersMemoryFileClass)
  Protected result.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = *this\usedFile
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a byte from the memory File object.
Procedure.b memoryFileClass_ReadByte(*this._membersMemoryFileClass)
  Protected result.b, *ptr.BYTE
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer < *this\usedFile
    *ptr = *this\FileBase + *this\filePointer
    result = *ptr\b
    *this\filePointer + SizeOf(BYTE)
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a characters from the memory File object.
Procedure.c memoryFileClass_ReadCharacter(*this._membersMemoryFileClass)
  Protected result.c, *ptr.CHARACTER
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer <= *this\usedFile - SizeOf(CHARACTER)
    *ptr = *this\FileBase + *this\filePointer
    result = *ptr\c
    *this\filePointer + SizeOf(CHARACTER)
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a block of data from the memory File object.
;Returns the number of bytes actually read.
Procedure.i memoryFileClass_ReadData(*this._membersMemoryFileClass, memoryBuffer, lengthToRead)
  Protected result
  If memoryBuffer And lengthToRead > 0
    If *this\mutex
      LockMutex(*this\mutex)
    EndIf
    If *this\filePointer > *this\usedFile - lengthToRead
      lengthToRead = *this\usedFile - *this\filePointer
    EndIf
    If lengthToRead > 0
      CopyMemory(*this\FileBase + *this\filePointer, memoryBuffer, lengthToRead)
      *this\filePointer + lengthToRead
      result = lengthToRead
    EndIf
    If *this\mutex
      UnlockMutex(*this\mutex)
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a double from the memory File object.
Procedure.d memoryFileClass_ReadDouble(*this._membersMemoryFileClass)
  Protected result.d, *ptr.DOUBLE
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer <= *this\usedFile - SizeOf(DOUBLE)
    *ptr = *this\FileBase + *this\filePointer
    result = *ptr\d
    *this\filePointer + SizeOf(DOUBLE)
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a float from the memory File object.
Procedure.f memoryFileClass_ReadFloat(*this._membersMemoryFileClass)
  Protected result.f, *ptr.FLOAT
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer <= *this\usedFile - SizeOf(FLOAT)
    *ptr = *this\FileBase + *this\filePointer
    result = *ptr\f
    *this\filePointer + SizeOf(FLOAT)
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads an integer from the memory File object.
Procedure.i memoryFileClass_ReadInteger(*this._membersMemoryFileClass)
  Protected result.i, *ptr.INTEGER
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer <= *this\usedFile - SizeOf(INTEGER)
    *ptr = *this\FileBase + *this\filePointer
    result = *ptr\i
    *this\filePointer + SizeOf(INTEGER)
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a long from the memory File object.
Procedure.l memoryFileClass_ReadLong(*this._membersMemoryFileClass)
  Protected result.l, *ptr.LONG
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer <= *this\usedFile - SizeOf(LONG)
    *ptr = *this\FileBase + *this\filePointer
    result = *ptr\l
    *this\filePointer + SizeOf(LONG)
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a quad from the memory File object.
Procedure.q memoryFileClass_ReadQuad(*this._membersMemoryFileClass)
  Protected result.q, *ptr.QUAD
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer <= *this\usedFile - SizeOf(QUAD)
    *ptr = *this\FileBase + *this\filePointer
    result = *ptr\q
    *this\filePointer + SizeOf(QUAD)
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a string from the memory File object.
Procedure.s memoryFileClass_ReadString(*this._membersMemoryFileClass, format=0)
  Protected result$, *ptr.ASCII, byteLength.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  ;Determine the format of the string to read.
    If format <> #PB_Ascii And format <> #PB_UTF8 And format <> #PB_Unicode
      CompilerIf #PB_Compiler_Unicode
        format = #PB_Unicode
      CompilerElse
        format = #PB_Ascii
      CompilerEndIf
    EndIf
  byteLength = StringByteLength("a", format) ;The 'a' takes the place of the terminating null(s).
  If *this\filePointer <= *this\usedFile - byteLength
    *ptr = *this\FileBase + *this\filePointer
    result$ = PeekS(*ptr, -1, format)
    ;We need to determine the number of bytes to advance the file pointer by. Complicated by the fact a memory buffer could be created
    ;containing UTF-8 strings by a Unicode program but read by an Ascii program. A typical example is a network program. This means that
    ;StringByteLength() cannot be used.
      While *ptr\a
        *ptr+1
        byteLength + 1
      Wend
    *this\filePointer + byteLength
    If *this\filePointer > *this\usedFile
      *this\filePointer = *this\usedFile
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method reads a word from the memory File object.
Procedure.w memoryFileClass_ReadWord(*this._membersMemoryFileClass)
  Protected result.w, *ptr.WORD
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If *this\filePointer <= *this\usedFile - SizeOf(WORD)
    *ptr = *this\FileBase + *this\filePointer
    result = *ptr\w
    *this\filePointer + SizeOf(WORD)
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method resets the memory File object.
;Returns zero if error. In which case the existing memory file is left unchanged.
Procedure.i memoryFileClass_Reset(*this._membersMemoryFileClass, newInitialSize=-1)
  Protected result.i, i.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  If newInitialSize <= 0
    newInitialSize = *this\initialSize
  EndIf
  result = ReAllocateMemory(*this\fileBase, newInitialSize)
  If result
    With *this  
      \initialSize = newInitialSize
      \fileBase = result
      \fileSize = \initialSize
      \pageSize = \initialPageSize
      \usedFile = 0
      \filePointer = 0
      \blnErrorReported = 0
    EndWith
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method truncates the memory File object.
Procedure memoryFileClass_SetPageSize(*this._membersMemoryFileClass, pageSize)
  If pageSize > 0
    If *this\mutex
      LockMutex(*this\mutex)
    EndIf
    *this\pageSize = pageSize
    If *this\mutex
      UnlockMutex(*this\mutex)
    EndIf
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method truncates the memory File object.
Procedure memoryFileClass_Truncate(*this._membersMemoryFileClass)
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  *this\usedFile = *this\filePointer
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a byte to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteByte(*this._membersMemoryFileClass, byte.b)
  Protected result.i, *ptr.BYTE
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = memoryFileClass_AllocateExtraBytes(*this, SizeOf(BYTE))
  If result
    *ptr = *this\FileBase + *this\filePointer
    *ptr\b = byte
    *this\filePointer+SizeOf(BYTE)
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a character to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteCharacter(*this._membersMemoryFileClass, char.c)
  Protected result.i, *ptr.CHARACTER
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = memoryFileClass_AllocateExtraBytes(*this, SizeOf(CHARACTER))
  If result
    *ptr = *this\FileBase + *this\filePointer
    *ptr\c = char
    *this\filePointer+SizeOf(CHARACTER)
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a memory block to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteData(*this._membersMemoryFileClass, memoryBuffer, length)
  Protected result.i, *ptr.BYTE
  If memoryBuffer And length > 0
    If *this\mutex
      LockMutex(*this\mutex)
    EndIf
    result = memoryFileClass_AllocateExtraBytes(*this, length)
    If result
      *ptr = *this\FileBase + *this\filePointer
      CopyMemory(memoryBuffer, *this\FileBase + *this\filePointer, length)
      *this\filePointer+length
      If *this\filePointer > *this\usedFile
        *this\usedFile = *this\filePointer
      EndIf
    EndIf
    If *this\mutex
      UnlockMutex(*this\mutex)
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a double to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteDouble(*this._membersMemoryFileClass, double.d)
  Protected result.i, *ptr.DOUBLE
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = memoryFileClass_AllocateExtraBytes(*this, SizeOf(DOUBLE))
  If result
    *ptr = *this\FileBase + *this\filePointer
    *ptr\d = double
    *this\filePointer+SizeOf(DOUBLE)
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a float to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteFloat(*this._membersMemoryFileClass, float.f)
  Protected result.i, *ptr.FLOAT
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = memoryFileClass_AllocateExtraBytes(*this, SizeOf(FLOAT))
  If result
    *ptr = *this\FileBase + *this\filePointer
    *ptr\f = float
    *this\filePointer+SizeOf(FLOAT)
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes an integer to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteInteger(*this._membersMemoryFileClass, Integer.i)
  Protected result.i, *ptr.INTEGER
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = memoryFileClass_AllocateExtraBytes(*this, SizeOf(INTEGER))
  If result
    *ptr = *this\FileBase + *this\filePointer
    *ptr\i = integer
    *this\filePointer+SizeOf(INTEGER)
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a long to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteLong(*this._membersMemoryFileClass, long.l)
  Protected result.i, *ptr.LONG
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = memoryFileClass_AllocateExtraBytes(*this, SizeOf(LONG))
  If result
    *ptr = *this\FileBase + *this\filePointer
    *ptr\l = long
    *this\filePointer+SizeOf(LONG)
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a quad to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteQuad(*this._membersMemoryFileClass, quad.q)
  Protected result.i, *ptr.QUAD
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = memoryFileClass_AllocateExtraBytes(*this, SizeOf(QUAD))
  If result
    *ptr = *this\FileBase + *this\filePointer
    *ptr\q = quad
    *this\filePointer+SizeOf(QUAD)
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a string to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteString(*this._membersMemoryFileClass, text$, format=0)
  Protected result.i, ptr.i, byteLength.i
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  ;Need to determine how many bytes the string (+ terminator) will occupy.
    If format <> #PB_Ascii And format <> #PB_UTF8 And format <> #PB_Unicode
      CompilerIf #PB_Compiler_Unicode
        format = #PB_Unicode
      CompilerElse
        format = #PB_Ascii
      CompilerEndIf
    EndIf
    byteLength = StringByteLength(text$+"a", format) ;The 'a' takes the place of the terminating null(s).
  result = memoryFileClass_AllocateExtraBytes(*this, byteLength)
  If result
    ptr = *this\FileBase + *this\filePointer
    PokeS(ptr, text$,-1,format)
    *this\filePointer+byteLength
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following method writes a word to the memory File object.
;Returns zero if an error.
Procedure.i memoryFileClass_WriteWord(*this._membersMemoryFileClass, word.w)
  Protected result.i, *ptr.WORD
  If *this\mutex
    LockMutex(*this\mutex)
  EndIf
  result = memoryFileClass_AllocateExtraBytes(*this, SizeOf(WORD))
  If result
    *ptr = *this\FileBase + *this\filePointer
    *ptr\w = word
    *this\filePointer+SizeOf(WORD)
    If *this\filePointer > *this\usedFile
      *this\usedFile = *this\filePointer
    EndIf
  EndIf
  If *this\mutex
    UnlockMutex(*this\mutex)
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following function decides if extra memory is required in order to accommodate the specified additional number of bytes.
;If so, an attempt is made to allocate the extra memory etc.
;Returns zero if an error.
Procedure.i memoryFileClass_AllocateExtraBytes(*this._membersMemoryFileClass, numBytes.i)
  Protected result.i = 1
  Protected extraPages.i, memBytes.i, mem.i
  extraPages = *this\filePointer + numBytes - *this\FileSize
  If extraPages > 0
    ;Only proceed if the memory file object has a variable buffer.
    If *this\flags & #MEMORYFILE_VARIABLEBUFFER
      ;How many extra pages of memory do we require?
        extraPages + *this\pageSize - 1
        extraPages / *this\pageSize
      ;How many bytes is this in total that we need?
        memBytes = *this\FileSize + extraPages**this\pageSize
      ;Attempt to reallocate the File memory.
        result = ReAllocateMemory(*this\FileBase, memBytes)
        If result
          *this\FileBase = result
          *this\FileSize = memBytes
          result = 1
        Else
          *this\blnErrorReported = #True
        EndIf
    Else
      *this\blnErrorReported = #True
      result = 0    
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////



;-VIRTUAL TABLES.

DataSection 
  VTable_MemoryFileClass: 
    Data.i @memoryFileClass_CheckForErrors()
    Data.i @memoryFileClass_ClearErrorFlag()
    Data.i @memoryFileClass_Destroy() 
    Data.i @memoryFileClass_Eof()
    Data.i @memoryFileClass_FileSeek()
    Data.i @memoryFileClass_GetFileMemoryBase() 
    Data.i @memoryFileClass_GetPageSize()
    Data.i @memoryFileClass_Loc()
    Data.i @memoryFileClass_Lof()
    Data.i @memoryFileClass_ReadByte()
    Data.i @memoryFileClass_ReadCharacter()
    Data.i @memoryFileClass_ReadData()
    Data.i @memoryFileClass_ReadDouble()
    Data.i @memoryFileClass_ReadFloat()
    Data.i @memoryFileClass_ReadInteger()
    Data.i @memoryFileClass_ReadLong()
    Data.i @memoryFileClass_ReadQuad()
    Data.i @memoryFileClass_ReadString()
    Data.i @memoryFileClass_ReadWord()
    Data.i @memoryFileClass_Reset()
    Data.i @memoryFileClass_SetPageSize()
    Data.i @memoryFileClass_Truncate()
    Data.i @memoryFileClass_WriteByte()
    Data.i @memoryFileClass_WriteCharacter()
    Data.i @memoryFileClass_WriteData()
    Data.i @memoryFileClass_WriteDouble()
    Data.i @memoryFileClass_WriteFloat()
    Data.i @memoryFileClass_WriteInteger()
    Data.i @memoryFileClass_WriteLong()
    Data.i @memoryFileClass_WriteQuad()
    Data.i @memoryFileClass_WriteString()
    Data.i @memoryFileClass_WriteWord()
EndDataSection 

DisableExplicit

CompilerEndIf

; IDE Options = PureBasic 5.62 beta 1 (Windows - x64)
; Folding = -------