﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Classes module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;     i)    minScript classes are user defined types. They are reference counted (not all user defined types need use reference counting).
;
;     ii)   A class can only be instantiated from within a script. 
;
;NOTE to self.
;=============
;     Take care to distinguish *class\code and *this\sCode. The former represents the code object in which the class definition is housed whilst
;     the latter represents the code object in which the particular instance is housed.
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES

  Declare.i minScript_RunTime_NewClass(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariant, *ret.minScriptVariant)
  Declare minScriptClasses_GCE(codeObject, uMsg)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS

;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-STRUCTURES

  ;The following structure represents each object property and are held in a map.
    Structure st_minScriptObjectVariantClass_ClassProperty
      isPrivateProperty.i
      property.minScriptVariant
    EndStructure

  ;The following is the class template for the minScriptObjectVariant class representing our classes.
    Structure st_minScriptObjectVariantClass_Class
      *vTable
      refCount.i        ;When this hits zero, the object is removed.
      *sCode.st_minScriptCodeClass  ;The original code object used to instantiate the class.
      *class.st_minScriptClass
      ;Property fields.
        Map properties.st_minScriptObjectVariantClass_ClassProperty() 
      ;The following allows a class method to retrieve the number of parameters passed to a method (inc class constructor).
        numParamsPassedToMethod.i
      ;The following allows us to track whether a property or some such is being accessed from within a class method etc.
      ;If so, then we do not call an over-riden property get/set.
        blnIsInsideMethod.i 
      ;The following is used to prevent the object being destroyed whilst it is already being destroyed through a re-entrant \Clear() method and a zero reference count.
        blnBeingDestroyed.i
    EndStructure

  ;The following holds the information required to administer the class GC extension.
    Structure st_minScriptClass_GCE
      threadMutex.i
      Map *classObjects.st_minScriptObjectVariantClass_Class()  
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBALS

  Global minScriptClassesGCE_globals.st_minScriptClass_GCE
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
;The following class init function is called at startup to allow minScript to register a suitable run-time function etc.
Procedure.i minScript_InitClassTypes()
  Protected result = #minScript_OKAY
  ;Register a run-time function used to create new instances of array objects.
  ;This particular function is required to return an instance of a minScriptObjectVariant interface.
    result = minScript_AddRuntimeFunction("NEW", #minScript_STANDARDRUNTIME_NEWCLASS, @minScript_RunTime_NewClass())
  If result = #minScript_OKAY
    ;Create a mutex to protect access to the GCE data if appropriate.
      CompilerIf #PB_Compiler_Thread
        minScriptClassesGCE_globals\threadMutex = CreateMutex()
        If minScriptClassesGCE_globals\threadMutex = 0
          result = #minScript_OUTOFMEMORY
        EndIf
      CompilerEndIf
    If result = #minScript_OKAY
      ;Add appropriate GC extensions.
        minScript_AddGCExtension(@minScriptClasses_GCE())
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following helper function attempts to find a class map entry given a class name. 
;It first searches the given code object before then looking to the global library object.
Procedure.i minScriptClasses_HELPER_FindClass(*code.st_minScriptCodeClass, class$)
  Protected *class.st_minScriptClass, *gcl.st_minScriptCodeClass
  ;First check the given code object.
    *class = FindMapElement(*code\classes(), class$)
  If *class = 0
    ;Check the global library object.  
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScript_Globals\threadMutex)
      CompilerEndIf
      *gcl = minScriptGlobalClassLib
      *class = FindMapElement(*gcl\classes(), class$)
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScript_Globals\threadMutex)
      CompilerEndIf
  EndIf
  ProcedureReturn *class
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following helper function prepares the stack for a method invokation.
;We add as many NULL elements to the variant stack as required. Proceeding manually allows for the greatest speed.
;In fact since unused elements on the stack are always 'empty', we can just move the stack pointer.
Procedure.i minScriptClasses_HELPER_PrepareStack(*execution.st_minScriptCodeExecutionClass, numParamsAlreadyPassed)
  Protected result = #minScript_OKAY, numParamsToPush
  numParamsToPush = #minScript_MAXNUMFUNCTIONPARAMETERS - numParamsAlreadyPassed
  If numParamsToPush > 0
    If *execution\numElementsOnVariantStack + #minScript_MAXNUMFUNCTIONPARAMETERS - numParamsAlreadyPassed <= #minScript_VARIANTSTACKSIZE
      *execution\ESP + SizeOf(minScriptVariant) * numParamsToPush
      *execution\numElementsOnVariantStack + numParamsToPush
    Else
      result = #minScript_VARIANTSTACKOVERFLOW 
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our "NEWCLASS" runtime function used to initialise a class object.
Procedure.i minScript_RunTime_NewClass(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *this.st_minScriptObjectVariantClass_Class, *code.st_minScriptCodeClass, *class.st_minScriptClass, class$
  Protected methodName$, *function.st_minScriptFunction, t1, blnIsInsideMethod
  If numParamsPassed > 0
    If *params\index[0]\vt = #minScript_STRING
      *code = codeObject ;Allows us to access the fields of the underlying code object.
      If *params\index[0]\sPtr
        class$ = LCase(PeekS(*params\index[0]\sPtr))
      EndIf
      If class$
        ;We need to locate the appropriate class.
        *class = minScriptClasses_HELPER_FindClass(codeObject, class$)
        If *class
          ;Allocate memory for a st_minScriptObjectVariantClass_Class structure.
            *this = AllocateStructure(st_minScriptObjectVariantClass_Class)
          If *this
            *this\vTable = ?VTable_minScriptObjectVariantClass_Class
            ;Add the various properties.
              CompilerIf #PB_Compiler_Thread
                LockMutex(minScript_Globals\threadMutex)
              CompilerEndIf
              ForEach *class\propertyInfo()
                If AddMapElement(*this\properties(), MapKey(*class\propertyInfo())) = 0
                  result = #minScript_OUTOFMEMORY
                  Break
                EndIf
                *this\properties()\isPrivateProperty = *class\propertyInfo()\isPrivateProperty
                *this\properties()\property\vt = *class\propertyInfo()\vt
              Next
              CompilerIf #PB_Compiler_Thread
                UnlockMutex(minScript_Globals\threadMutex)
              CompilerEndIf
            If result = #minScript_OKAY
              *this\refCount = 1
              *this\sCode = codeObject
              *this\class = *class
              ;Call any constructor only if we are not already executing the constructor.
            CompilerIf #PB_Compiler_Thread
              If GetThreadStorageValue(*class\tsiConstructorActive) = #False
            CompilerElse              
              If *class\tsiConstructorActive = #False
            CompilerEndIf
                methodName$ = class$ + "@" + class$
                *function = minScriptCodeClass_HELPER_GetFunctionFromName(methodName$, *class\module)
                If *function
                  If numParamsPassed <= #minScript_MAXNUMFUNCTIONPARAMETERS
                    ;Need to set the parameters up carefully.
                      ;We need to make sure there are the expected number of parameters on the stack.
                        result = minScriptClasses_HELPER_PrepareStack(*code\execution, numParamsPassed)
                    If result = #minScript_OKAY
                      ;Now replace the class name string parameter with 'this'.
                        minScript_StringFree(*params\index[0]\sPtr)
                        *params\index[0]\vt = *this\class\classVT
                        *params\index[0]\refPtr = *this
                        *this\refCount + 1  ;This will be balanced when minScript tidies up the stack frame for this call.
                      blnIsInsideMethod = *this\blnIsInsideMethod
                      *this\blnIsInsideMethod = #True
                      *this\numParamsPassedToMethod = numParamsPassed-1
                      CompilerIf #PB_Compiler_Thread
                        SetThreadStorageValue(*class\tsiConstructorActive, #True)
                      CompilerElse              
                        *class\tsiConstructorActive = #True
                      CompilerEndIf
                      ;Invoke the constructor method.
                        result = *code\execution\Execute(*function)
                      CompilerIf #PB_Compiler_Thread
                        SetThreadStorageValue(*class\tsiConstructorActive, #False)
                      CompilerElse              
                        *class\tsiConstructorActive = #False
                      CompilerEndIf
                      *this\blnIsInsideMethod = blnIsInsideMethod
                      *this\numParamsPassedToMethod = 0
                      ;Clear the EAX register in case the constructor returned some value - which we discard.
                        t1 = minScript_VarClear(*ret)
                        If result = #minScript_OKAY
                          result = t1
                        EndIf
                    EndIf 
                  Else
                    result = #minScript_TOOMANYPARAMETERS                  
                  EndIf
                EndIf          
              EndIf
              If result = #minScript_OKAY
                ;Add to the classes GC extension.
                  CompilerIf #PB_Compiler_Thread
                    LockMutex(minScriptClassesGCE_globals\threadMutex)
                  CompilerEndIf
                    If AddMapElement(minScriptClassesGCE_globals\classObjects(), Str(*this))
                      minScriptClassesGCE_globals\classObjects() = *this
                    Else
                      result = #minScript_OUTOFMEMORY
                    EndIf
                  CompilerIf #PB_Compiler_Thread
                    UnlockMutex(minScriptClassesGCE_globals\threadMutex)
                  CompilerEndIf
                If result = #minScript_OKAY                      
                  ;Set the *ret fields.
                  *ret\vt = *this\class\classVT
                  *ret\refPtr = *this
                Else
                  FreeStructure(*this)            
                EndIf
              Else
                FreeStructure(*this)            
              EndIf
            Else
              FreeStructure(*this)            
            EndIf
          Else
            result = #minScript_OUTOFMEMORY      
          EndIf
        Else
          result = #minScript_CLASSNOTFOUND      
        EndIf
      Else
        result = #minScript_INVALIDPARAMETER
      EndIf
    Else
      result = #minScript_INVALIDDATATYPE
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS  
  EndIf
  ;We will clear #minScript_OUTOFMEMORY errors so that scripts can determine for themselves how to proceed. A script can always use the THROWERROR() run-time function.
  If result = #minScript_OUTOFMEMORY
    result = #minScript_OKAY
  EndIf  
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-GC Extension.

;/////////////////////////////////////////////////////////////////////////////////
;The following GC extension disposes of any class objects unaccounted for.
Procedure minScriptClasses_GCE(codeObject, uMsg)
  Select uMsg
    Case #minScript_GCEXTENSION_CODEOBJECTHASBEENCREATED
      ;Not needed.
    Default ; #minScript_GCEXTENSION_CODEOBJECTHASBEENRESET or #minScript_GCEXTENSION_CODEOBJECTHASBEENDELETED
      ;We simply loop through all outstanding class objects and free those belonging to the given code object.
        CompilerIf #PB_Compiler_Thread
          LockMutex(minScriptClassesGCE_globals\threadMutex)
        CompilerEndIf
          ForEach minScriptClassesGCE_globals\classObjects()
            If minScriptClassesGCE_globals\classObjects()\sCode = codeObject
              ;We forcibly delete the class object by clearing all properties.
              ;We must avoid clearing any reference type properties.
                ForEach minScriptClassesGCE_globals\classObjects()\properties()
                  If minScriptClassesGCE_globals\classObjects()\properties()\property\vt & #minScript_REFERENCE = 0
                    minScript_VarClear(minScriptClassesGCE_globals\classObjects()\properties()\property)
                  EndIf
                Next
              FreeStructure(minScriptClassesGCE_globals\classObjects())
              DeleteMapElement(minScriptClassesGCE_globals\classObjects())
            EndIf
          Next
        CompilerIf #PB_Compiler_Thread
          UnlockMutex(minScriptClassesGCE_globals\threadMutex)
        CompilerEndIf
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-minScriptObjectVariant CLASS METHODS.
;These 3 methods are required for scripts to be able to make use of our classes.

;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \Clear() method.
Procedure minScriptObjectVariantClass_Class_Clear(*this.st_minScriptObjectVariantClass_Class)
  Protected result = #minScript_OKAY, methodName$, *function.st_minScriptFunction, t1, *sExecution.st_minScriptCodeExecutionClass
  Protected var.minScriptVariant, blnIsInsideMethod
  *this\refCount - 1
  If *this\refCount = 0 And *this\blnBeingDestroyed = #False
    ;Ensure we return any current run-time error.
      result = *this\sCode\lastErrorCode
    *this\blnBeingDestroyed = #True
    ;Call any destructor which may be defined. We must only do this if there is currently a script running.
    If *this\sCode\numCodeExecutions And result = #minScript_OKAY
      methodName$ = *this\class\name$ + "@~" + *this\class\name$
      *function = minScriptCodeClass_HELPER_GetFunctionFromName(methodName$, *this\class\module)
    CompilerIf #PB_Compiler_Thread
      If *function  And GetThreadStorageValue(*this\class\tsiDestructorActive) = #False
    CompilerElse              
      If *function And *this\class\tsiDestructorActive = #False
    CompilerEndIf
        *sExecution = *this\sCode\execution
      ;Need to set the parameters up carefully.
      ;First a 'this' parameter. We do this manually, as opposed to calling the \MS_OP_PUSHVAR() method.
        If *sExecution\numElementsOnVariantStack < #minScript_VARIANTSTACKSIZE
          *sExecution\ESP + SizeOf(minScriptVariant)
          *sExecution\numElementsOnVariantStack + 1
          result = minScriptClasses_HELPER_PrepareStack(*sExecution, 1)
          If result = #minScript_OKAY
            ;Set the 'this' parameter.
              *sExecution\ESP\vt = *this\class\classVT
              *sExecution\ESP\refPtr = *this
              *this\refCount = 1 ;Accounts for the temporary 'this' pointer.
            ;Preserve and empty the EAX register.
              var\vt = *this\sCode\EAX\vt
              var\q = *this\sCode\EAX\q              
              *this\sCode\EAX\vt = #minscript_EMPTY
              *this\sCode\EAX\q = 0
            ;Invoke the destructor method.
              blnIsInsideMethod = *this\blnIsInsideMethod ;Not really needed.
              *this\blnIsInsideMethod = #True
              *this\numParamsPassedToMethod = 0
              CompilerIf #PB_Compiler_Thread
                SetThreadStorageValue(*this\class\tsiDestructorActive, #True)
              CompilerElse              
                *this\class\tsiDestructorActive = #True
              CompilerEndIf
              result = *this\sCode\execution\Execute(*function)
              CompilerIf #PB_Compiler_Thread
                SetThreadStorageValue(*this\class\tsiDestructorActive, #False)
              CompilerElse              
                *this\class\tsiDestructorActive = #False
              CompilerEndIf
              *this\blnIsInsideMethod = blnIsInsideMethod
            ;Reinstate EAX.
              t1 = minScript_VarClear(*this\sCode\EAX)
              *this\sCode\EAX\vt = var\vt
              *this\sCode\EAX\q = var\q
            If result = #minScript_OKAY
              result = t1
            EndIf              
          EndIf
        Else
          result = #minScript_VARIANTSTACKOVERFLOW 
        EndIf
      EndIf
    EndIf
    ;Unless the destructor has increased the reference count (by copying the 'this' object to a module variable for example), we remove all properties. Order is not important.
      If *this\refCount = 0
        ForEach *this\properties()
          t1 = minScript_VarClear(*this\properties()\property)
          If result = #minScript_OKAY
            result = t1
          EndIf
        Next
        FreeStructure(*this)
        ;Remove from the classes GC extension.
          CompilerIf #PB_Compiler_Thread
            LockMutex(minScriptClassesGCE_globals\threadMutex)
          CompilerEndIf
          DeleteMapElement(minScriptClassesGCE_globals\classObjects(), Str(*this))
          CompilerIf #PB_Compiler_Thread
            UnlockMutex(minScriptClassesGCE_globals\threadMutex)
          CompilerEndIf
      Else
        *this\blnBeingDestroyed = #False
      EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following is our implementation of the prerequisite \CopyToVariant() method.
Procedure.i minScriptObjectVariantClass_Class_CopyToVariant(*this.st_minScriptObjectVariantClass_Class, *varDest.minScriptVariant)
  Protected result = #minScript_OKAY
  *varDest\vt = *this\class\classVT
  *varDest\refPtr = *this
  *this\refCount + 1
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScriptObjectVariantClass_Class_Invoke(*this.st_minScriptObjectVariantClass_Class, *code.st_minScriptCodeClass, methodName$, invokeType, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, blnIsInsideMethod, *function.st_minScriptFunction
  Protected *ptrProperty.st_minScriptObjectVariantClass_ClassProperty, previousNumParamsPassed
  ;If we are inside a method then we allow access to the numParamsPassedToMethod property.
  If methodName$ = "numparamspassed"
    If invokeType & #minScript_METHOD_GET
      *ret\vt = #minScript_INTEGER     
      If *this\blnIsInsideMethod
        *ret\i = *this\numParamsPassedToMethod
      Else
        *ret\i = 0
      EndIf
    ElseIf invokeType = #minScript_METHOD_INVOKE
      result = #minScript_METHODORPROPERTYNOTSUPPORTED
    Else
      result = #minScript_PROPERTYISREADONLY
    EndIf
  Else
    previousNumParamsPassed = *this\numParamsPassedToMethod
    *this\numParamsPassedToMethod = numParamsPassed-1
    ;Record that we are inside a method so that we only call any overriding property get/sets if invoked from outside a method.
      blnIsInsideMethod = *this\blnIsInsideMethod
      *this\blnIsInsideMethod = #True
    If invokeType = #minScript_METHOD_INVOKE
      ;Decorate the method name.
        methodName$ = *this\class\name$ + "@" + methodName$
      ;Locate the method function.
        *function = minScriptCodeClass_HELPER_GetFunctionFromName(methodName$, *this\class\module)
      If *function
        ;We need to make sure there are the expected number of parameters on the stack
          result = minScriptClasses_HELPER_PrepareStack(*code\execution, numParamsPassed)
        If result = #minScript_OKAY
          ;Invoke the property method.
            result = *code\execution\Execute(*function)
        EndIf
      Else
        result = #minScript_METHODORPROPERTYNOTSUPPORTED
      EndIf
    Else ; A get or a set.
      If Left(methodName$, 1) <> "@"
        ;Locate the property. Will only succeed if the property exists or is not a virtual one.
          If *this\blnBeingDestroyed
            PushMapPosition(*this\properties())
          EndIf
          *ptrProperty = FindMapElement(*this\properties(), methodName$)
          If *this\blnBeingDestroyed
            PopMapPosition(*this\properties())
          EndIf
        ;Decorate the method name.
          If invokeType & #minScript_METHOD_GET
            methodName$ = *this\class\name$ + "@get_" + methodName$
          Else
            methodName$ = *this\class\name$ + "@set_" + methodName$
          EndIf
        ;Locate any method being used to over-ride the property.
          *function = minScriptCodeClass_HELPER_GetFunctionFromName(methodName$, *this\class\module)
        ;Are we to invoke any over-ridden property?
        If blnIsInsideMethod = #False And *function
          ;We need to make sure there are the expected number of parameters on the stack.
            result = minScriptClasses_HELPER_PrepareStack(*code\execution, numParamsPassed)
          If result = #minScript_OKAY
            ;Invoke the property method. *ret points to EAX and so this will be set automatically.
              result = *code\execution\Execute(*function)
            ;Promote the return type if the property is not a virtual one.
            If result = #minScript_OKAY And *ptrProperty And (*ptrProperty\property\vt&#minScript_VARIANT = 0)
              result = minScript_VarChangeType(*ret, *ptrProperty\property\vt)
            EndIf
          EndIf
        Else ;We simply retrieve/set the underlying property - if one exists and we are allowed access.
          If *ptrProperty And (*ptrProperty\isPrivateProperty = #False Or blnIsInsideMethod = #True)
            If invokeType & #minScript_METHOD_GET
              ;We should only have a single parameter ('this').
              If numParamsPassed = 1
                ;We need to copy the property.
                  result = minScript_VarCopy(*ret, *ptrProperty\property, #minScript_VARDONOTFREEDESTINATIONVARIABLE)
              Else
                result = #minScript_INCORRECTNUMBEROFPARAMETERS
              EndIf
            Else ;set.
              If *this\blnBeingDestroyed = #False ;We don't return an error here.
                ;We should only have two parameters ('this', value).
                If numParamsPassed = 2
                  ;Change the type of the given parameter if appropriate.
                  If *ptrProperty\property\vt & #minScript_VARIANT
                    result = minScript_VarClear(*ptrProperty\property, #minScript_VARDONOTCLEARVTFIELD)
                    *ptrProperty\property\vt = *params\index[1]\vt|#minScript_VARIANT
                    *ptrProperty\property\q = *params\index[1]\q
                    ;Clear the given parameter. No need to free the variant as we have copied it.
                      *params\index[1]\vt = #minScript_EMPTY
                      *params\index[1]\q = 0
                  Else
                    result = minScript_VarChangeType(*params\index[1], *ptrProperty\property\vt)
                    If result = #minScript_OKAY
                      result = minScript_VarClear(*ptrProperty\property, #minScript_VARDONOTCLEARVTFIELD)
                      ;Copy the value.
                        *ptrProperty\property\q = *params\index[1]\q
                      ;Clear the given parameter. No need to free the variant as we have copied it.
                        *params\index[1]\vt = #minScript_EMPTY
                        *params\index[1]\q = 0
                    EndIf
                  EndIf
                Else
                  result = #minScript_INCORRECTNUMBEROFPARAMETERS
                EndIf
              EndIf
            EndIf
          Else
            result = #minScript_METHODORPROPERTYNOTSUPPORTED
          EndIf
        EndIf
      Else ;'@'
        ;Simply return the address of the property.
        methodName$ = Right(methodName$, Len(methodName$)-1)
        ;Locate the property. Will only succeed if the property exists or is not a virtual one.
          If *this\blnBeingDestroyed
            PushMapPosition(*this\properties())
          EndIf
          *ptrProperty = FindMapElement(*this\properties(), methodName$)
          If *this\blnBeingDestroyed
            PopMapPosition(*this\properties())
          EndIf
        If *ptrProperty And (*ptrProperty\isPrivateProperty = #False Or blnIsInsideMethod = #True)
          ;We should only have a single parameter ('this').
          If numParamsPassed = 1
            *ret\vt = #minScript_INTEGER
            If *ptrProperty\property\vt&#minscript_VARIANTMASK = #minScript_STRING
              *ret\i = *ptrProperty\property\sPtr
            Else
              *ret\i = *ptrProperty\property
            EndIf
          Else
            result = #minScript_INCORRECTNUMBEROFPARAMETERS
          EndIf
        Else
          result = #minScript_METHODORPROPERTYNOTSUPPORTED
        EndIf
      EndIf
    EndIf
    *this\numParamsPassedToMethod = previousNumParamsPassed
    ;Reinstate the blnIsInsideMethod field
    *this\blnIsInsideMethod = blnIsInsideMethod
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


DataSection 
  VTable_minScriptObjectVariantClass_Class: 
    Data.i @minScriptObjectVariantClass_Class_Clear()
    Data.i @minScriptObjectVariantClass_Class_CopyToVariant()
    Data.i @minScriptObjectVariantClass_Class_Invoke()
EndDataSection



; IDE Options = PureBasic 5.62 (Windows - x86)
; ExecutableFormat = Shared dll
; CursorPosition = 333
; FirstLine = 323
; Folding = --
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode