﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Runtime Functions Module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;-DECLARES
  Declare minScript_RunTime_GCE(codeObject, uMsg)
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-CONSTANTS and STRUCTURES

  ;The following holds info on an individual resource (memory address or fileID etc.) so that the relevant GCE can dispose of it if needs be.
    Structure st_minScript_RunTime_GCE_Resource
      codeObject.i
      StructureUnion
        address.i       ;Memory allocations viz ALLOCATEMEMORY() etc.
        fileID.i        ;FileID viz OPENFILE() etc.
      EndStructureUnion
    EndStructure

  ;The following holds the information required to administer all GC extensions associated with the standard runtime.
    Structure st_minScript_RunTime_GCE
      threadMutex.i                 ;We use a single mutex for all runtime GC extensions.
      ;Lists should be fine for what should only ever be a relatively small number of resource allocations.
        List memoryAllocations.st_minScript_RunTime_GCE_Resource()  
        List fileIDs.st_minScript_RunTime_GCE_Resource()
    EndStructure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;-GLOBALS

  Global minScript_RunTime_GCE_globals.st_minScript_RunTime_GCE
;/////////////////////////////////////////////////////////////////////////////////


;-PUBLIC FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_AddRuntimeFunction(name$, fnID.l, address)
  Protected result = #minScript_OKAY, t1$, *ptrRuntimeFunction.st_minScriptRuntimeFunction, *ptr.INTEGER
  If name$ And fnID And address
    t1$ = LCase(name$)
    ;First check that the name does not match that of a minScript keyword.
    If (minScript_IsCompilerKeyword(t1$) = #False And t1$ <> "this")
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScript_Globals\threadMutex)
      CompilerEndIf
      ;Is there already a run-time function with this name?
      *ptrRuntimeFunction = FindMapElement(minScript_Globals\runtimeFunctionsTableByName(), t1$)
      If *ptrRuntimeFunction
        ;Only proceed if the fnID's match.
        If fnId = *ptrRuntimeFunction\fnID
          ;Check that we do not have an immutable compiler function. Standard run-time functions can be over-ridden.            
          If *ptrRuntimeFunction\fnID & #minScript_COMPILERRUNTIME_BASEID = #False
            *ptrRuntimeFunction\friendlyName$ = name$
            *ptrRuntimeFunction\address = address
          Else
            result = #minScript_ACCESSDENIED
          EndIf
        Else
          result = #minScript_FUNCTIONIDMISMATCH
        EndIf
      Else ;A completely new function.
        ;Check that the function name is valid.
        result = minScript_SymbolsIsSymbolNameValid(@t1$)
        If result = #minScript_OKAY
          ;Check that the fnID does not already exist.
          *ptr = FindMapElement(minScript_Globals\runtimeFunctionsTableByID(), Str(fnID))
          If *ptr = 0
            ;Attempt to add the 'new' function to the main run-time function tables.
            *ptrRuntimeFunction = AddMapElement(minScript_Globals\runtimeFunctionsTableByName(), t1$)
            If *ptrRuntimeFunction
              *ptrRuntimeFunction\friendlyName$ = name$
              *ptrRuntimeFunction\fnID = fnID
              *ptrRuntimeFunction\address = address
              *ptr = AddMapElement(minScript_Globals\runtimeFunctionsTableByID(), Str(fnID))
              If *ptr
                *ptr\i = *ptrRuntimeFunction
              Else
                DeleteMapElement(minScript_Globals\runtimeFunctionsTableByName())
                result = #minScript_OUTOFMEMORY
              EndIf
            Else
              result = #minScript_OUTOFMEMORY
            EndIf
          Else
            result = #minScript_FUNCTIONIDALREADYEXISTS
          EndIf
        EndIf
      EndIf
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScript_Globals\threadMutex)
      CompilerEndIf
    Else
      result = #minScript_SYMBOLNAMEMATCHESCOMPILERKEYWORD
    EndIf
  Else
    result = #minScript_INVALIDPARAMETER
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Excludes internal compiler functions.
;Returns a tab separated list of keywords.
Procedure.s minScript_GetListOfRuntimeFunctions()
  Protected name$, fnID
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ForEach minScript_Globals\runtimeFunctionsTableByName()
    ;Ignore internal compiler functions.
      If minScript_Globals\runtimeFunctionsTableByName()\fnID & #minScript_COMPILERRUNTIME_BASEID = 0
        name$ + minScript_Globals\runtimeFunctionsTableByName()\friendlyName$ + #TAB$
      EndIf
  Next
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn name$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;Returns the name of the required function (if found) and fills the given buffer with the address of the function.
Procedure.s minScript_GetRuntimeFunctionFromID(fnID.l, *ptrAddress.INTEGER=0)
  Protected name$, *ptr.INTEGER, *ptrRuntime.st_minScriptRuntimeFunction
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  *ptr = FindMapElement(minScript_Globals\runtimeFunctionsTableByID(), Str(fnID))
  If *ptr
    *ptrRuntime = *ptr\i
    name$ = *ptrRuntime\friendlyName$
    If *ptrAddress
      *ptrAddress\i = *ptrRuntime\address
    EndIf
  EndIf
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn name$
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Returns a minScript error code and fills the given buffers as appropriate.
Procedure.i minScript_GetRuntimeFunctionFromName(name$, *ptrfnID.INTEGER=0, *ptrAddress.INTEGER=0)
  Protected result=#minScript_OKAY, *ptrRuntime.st_minScriptRuntimeFunction
  CompilerIf #PB_Compiler_Thread
    LockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  *ptrRuntime = FindMapElement(minScript_Globals\runtimeFunctionsTableByName(), LCase(Trim(name$)))
  If *ptrRuntime
    If *ptrfnID
      *ptrfnID\i = *ptrRuntime\fnID
    EndIf
    If *ptrAddress
      *ptrAddress\i = *ptrRuntime\address
    EndIf
  Else
    result = #minScript_FUNCTIONNOTFOUND
  EndIf
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(minScript_Globals\threadMutex)
  CompilerEndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;-STANDARD RUNTIME


;/////////////////////////////////////////////////////////////////////////////////
;Some of our file run-time functions which take a single integer as a parameter.
Procedure.i minScript_RunTime_FILE1(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 1
    If *params\index[0]\vt = #minScript_INTEGER
      Select fnID
        Case #minScript_STANDARDRUNTIME_CLOSEFILE
          CloseFile(*params\index[0]\i)
          ;Remove from the GCE memory list.
          CompilerIf #PB_Compiler_Thread
            LockMutex(minScript_RunTime_GCE_globals\threadMutex)
          CompilerEndIf
            ForEach minScript_RunTime_GCE_globals\fileIDs()
              If minScript_RunTime_GCE_globals\fileIDs()\fileID = *params\index[0]\i
                DeleteElement(minScript_RunTime_GCE_globals\fileIDs())
                Break
              EndIf           
            Next
          CompilerIf #PB_Compiler_Thread
            UnlockMutex(minScript_RunTime_GCE_globals\threadMutex)
          CompilerEndIf
        Case #minScript_STANDARDRUNTIME_EOF
          *ret\vt = #minScript_INTEGER
          *ret\i = Eof(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_LOC
          *ret\vt = #minScript_INTEGER
          *ret\i = Loc(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_LOF
          *ret\vt = #minScript_INTEGER
          *ret\i = Lof(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_READBYTE
          *ret\vt = #minScript_INTEGER
          *ret\i = ReadByte(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_READDOUBLE
          *ret\vt = #minScript_REAL
          *ret\r = ReadDouble(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_READFLOAT
          *ret\vt = #minScript_REAL
          *ret\r = ReadFloat(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_READINTEGER
          *ret\vt = #minScript_INTEGER
          *ret\i = ReadInteger(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_READLONG
          *ret\vt = #minScript_INTEGER
          *ret\i = ReadLong(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_READSTRINGFORMAT
          *ret\vt = #minScript_INTEGER
          *ret\i = ReadStringFormat(*params\index[0]\i)
        Case #minScript_STANDARDRUNTIME_READWORD
          *ret\vt = #minScript_INTEGER
          *ret\i = ReadWord(*params\index[0]\i)
      EndSelect
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Some of our file run-time functions which take a single string as a parameter.
Procedure.i minScript_RunTime_FILE2(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrString.STRING
  If numParamsPassed = 1
    If *params\index[0]\vt = #minScript_STRING
      *ptrString = *params\index[0]
      Select fnID
        Case #minScript_STANDARDRUNTIME_FILESIZE
          *ret\vt = #minScript_INTEGER
          *ret\i = FileSize(*ptrString\s)
        Case #minScript_STANDARDRUNTIME_GETEXTENSIONPART
          result = minScript_VarSetString(*ret, GetExtensionPart(*ptrString\s), #False)
        Case #minScript_STANDARDRUNTIME_GETFILEPART
          result = minScript_VarSetString(*ret, GetFilePart(*ptrString\s), #False)
        Case #minScript_STANDARDRUNTIME_GETPATHPART
          result = minScript_VarSetString(*ret, GetPathPart(*ptrString\s), #False)
      EndSelect
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Some of our file run-time functions which take two parameters, the first being a fileID.
Procedure.i minScript_RunTime_FILE3(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 2
    If *params\index[0]\vt = #minScript_INTEGER
      Select fnID
        Case #minScript_STANDARDRUNTIME_WRITEBYTE
          minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
          *ret\vt = #minScript_INTEGER
          *ret\i = WriteByte(*params\index[0]\i, *params\index[1]\i)
        Case #minScript_STANDARDRUNTIME_WRITEDOUBLE
          minScript_VarChangeType(*params\index[1], #minScript_REAL)
          *ret\vt = #minScript_INTEGER
          *ret\i = WriteDouble(*params\index[0]\i, *params\index[1]\r)
        Case #minScript_STANDARDRUNTIME_WRITEFLOAT
          minScript_VarChangeType(*params\index[1], #minScript_REAL)
          *ret\vt = #minScript_INTEGER
          *ret\i = WriteFloat(*params\index[0]\i, *params\index[1]\r)
        Case #minScript_STANDARDRUNTIME_WRITEINTEGER
          minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
          *ret\vt = #minScript_INTEGER
          *ret\i = WriteInteger(*params\index[0]\i, *params\index[1]\i)
        Case #minScript_STANDARDRUNTIME_WRITELONG
          minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
          *ret\vt = #minScript_INTEGER
          *ret\i = WriteLong(*params\index[0]\i, *params\index[1]\i)
        Case #minScript_STANDARDRUNTIME_WRITEWORD
          minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
          *ret\vt = #minScript_INTEGER
          *ret\i = WriteWord(*params\index[0]\i, *params\index[1]\i)
        Case #minScript_STANDARDRUNTIME_WRITESTRINGFORMAT
          If *params\index[1]\vt = #minScript_INTEGER
            *ret\vt = #minScript_INTEGER
            *ret\i = WriteStringFormat(*params\index[0]\i, *params\index[1]\i)
          Else
            result = #minScript_INVALIDARGUMENT
          EndIf
      EndSelect
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Our file run-time functions dealing with data buffers. READDATA() and WRITEDATA().
Procedure.i minScript_RunTime_FILEDATA(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 3 
    If *params\index[0]\vt = #minScript_INTEGER And *params\index[1]\vt = #minScript_INTEGER And *params\index[2]\vt = #minScript_INTEGER And *params\index[1]\i And *params\index[2]\i > 0
      Select fnID
        Case #minScript_STANDARDRUNTIME_READDATA
          *ret\vt = #minScript_INTEGER
          *ret\i = ReadData(*params\index[0]\i, *params\index[1]\i, *params\index[2]\i)
      EndSelect
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_FINDSTRING(*code.st_minScriptCodeClass, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrSourceString.STRING, *ptrStringToFind.STRING, startPos=1, mode = #PB_String_CaseSensitive
  If numParamsPassed >= 2 And numParamsPassed <=4 
    ;Convert the first two arguments to strings.
      result = minScript_VarChangeType(*params\index[0], #minScript_STRING)
    If result = #minScript_OKAY
      result = minScript_VarChangeType(*params\index[1], #minScript_STRING)
      If result = #minScript_OKAY
        ;Sort out the default parameters.
          If numParamsPassed >= 3
            minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
            startPos = *params\index[2]\i
            If numParamsPassed = 4
              minScript_VarChangeType(*params\index[3], #minScript_INTEGER)
              mode = *params\index[3]\i
            EndIf
          EndIf    
        *ptrSourceString = *params\index[0]
        *ptrStringToFind = *params\index[1]
        *ret\vt = #minScript_INTEGER
        *ret\i = FindString(*ptrSourceString\s, *ptrStringToFind\s, startPos, mode)
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following handles the Left() and Right() standard run-time functions.
Procedure.i minScript_RunTime_LEFTRIGHT(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrString.STRING
  If numParamsPassed = 2
    ;Change the first parameter to a string.
      result = minScript_VarChangeType(*params\index[0], #minScript_STRING)
    If result = #minScript_OKAY
      *ptrString = *params\index[0]
      ;Change the second parameter to an integer.
        minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
      Select fnID
        Case #minScript_STANDARDRUNTIME_LEFT
          result = minScript_VarSetString(*ret, Left(*ptrString\s, *params\index[1]\i), #False)
        Case #minScript_STANDARDRUNTIME_RIGHT
          result = minScript_VarSetString(*ret, Right(*ptrString\s, *params\index[1]\i), #False)
      EndSelect
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_INTEGER2INTEGER(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 1
    ;Change the parameter to an integer.
      result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
    If result = #minScript_OKAY
      *ret\vt = #minScript_INTEGER
      Select fnID
        Case #minScript_STANDARDRUNTIME_SIZEOF
          If *params\index[0]\i & #minScript_REFERENCE
            *ret\i = SizeOf(INTEGER)
          Else
            *ret\i = 0
            Select *params\index[0]\i
              Case #minScript_INTEGER, #minScript_STRING
                *ret\i = SizeOf(INTEGER)      
              Case #minScript_CURRENCY
                *ret\i = SizeOf(QUAD)      
              Case #minScript_REAL
                *ret\i = SizeOf(DOUBLE)      
              Case #minScript_BYTE
                *ret\i = SizeOf(BYTE)      
              Case #minScript_WORD
                *ret\i = SizeOf(WORD)      
              Case #minScript_CHARACTER
                *ret\i = SizeOf(CHARACTER)      
              Case #minScript_LONG
                *ret\i = SizeOf(LONG)      
              Case #minScript_FLOAT
                *ret\i = SizeOf(FLOAT)      
              Case #minScript_VARIANT
                *ret\i = SizeOf(minScriptVariant)      
            EndSelect
          EndIf
        Case #minScript_STANDARDRUNTIME_ALLOCATEMEMORY
          If *params\index[0]\i > 0
            *ret\i = AllocateMemory(*params\index[0]\i)
            If *ret\i
              ;Add to the GCE memory list.
                CompilerIf #PB_Compiler_Thread
                  LockMutex(minScript_RunTime_GCE_globals\threadMutex)
                CompilerEndIf
                  If AddElement(minScript_RunTime_GCE_globals\memoryAllocations())
                    minScript_RunTime_GCE_globals\memoryAllocations()\codeObject = codeObject
                    minScript_RunTime_GCE_globals\memoryAllocations()\address = *ret\i
                  Else ;Should free the memory just in case.
                    FreeMemory(*ret\i)
                    *ret\i = 0
                  EndIf
                CompilerIf #PB_Compiler_Thread
                  UnlockMutex(minScript_RunTime_GCE_globals\threadMutex)
                CompilerEndIf
            EndIf
          EndIf
        Case #minScript_STANDARDRUNTIME_FREEMEMORY
          If *params\index[0]\i
            *ret\i = FreeMemory(*params\index[0]\i)
            ;Remove from the GCE memory list.
              CompilerIf #PB_Compiler_Thread
                LockMutex(minScript_RunTime_GCE_globals\threadMutex)
              CompilerEndIf
                ForEach minScript_RunTime_GCE_globals\memoryAllocations()
                  If minScript_RunTime_GCE_globals\memoryAllocations()\address = *params\index[0]\i
                    DeleteElement(minScript_RunTime_GCE_globals\memoryAllocations())
                    Break
                  EndIf           
                Next
              CompilerIf #PB_Compiler_Thread
                UnlockMutex(minScript_RunTime_GCE_globals\threadMutex)
              CompilerEndIf
         EndIf
        Case #minScript_STANDARDRUNTIME_THROWERROR
          result = *params\index[0]\i
      EndSelect
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following handles those standard run-time functions taking a single integer parameter and returning a string (e.g. "CHR").
Procedure.i minScript_RunTime_INTEGER2STRING(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrChar.CHARACTER
  If numParamsPassed = 1
    ;Change the parameter to an integer.
      result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
    If result = #minScript_OKAY
      Select fnID
        Case #minScript_STANDARDRUNTIME_CHR
          result = minScript_VarSetString(*ret, Chr(*params\index[0]\i), #False)
        Case #minScript_STANDARDRUNTIME_HEX
          result = minScript_VarSetString(*ret, Hex(*params\index[0]\i))
        Case #minScript_STANDARDRUNTIME_SPACE
          result = minScript_VarSetString(*ret, Space(*params\index[0]\i), #False)
      EndSelect
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_MID(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrString.STRING
  If numParamsPassed = 2 Or numParamsPassed = 3
    ;Change the first parameter to a string.
      result = minScript_VarChangeType(*params\index[0], #minScript_STRING)
    If result = #minScript_OKAY
      *ptrString = *params\index[0]
      ;Change the second parameter to an integer.
        minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
      If numParamsPassed = 2
        result = minScript_VarSetString(*ret, Mid(*ptrString\s, *params\index[1]\i), #False)
      Else
        ;Change any optional 3rd parameter to a string.
          minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
        result = minScript_VarSetString(*ret, Mid(*ptrString\s, *params\index[1]\i, *params\index[2]\i), #False)
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_NOTHING(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrString.STRING
  If numParamsPassed = 0
    ;The following 2 lines are not really required since *ret (EAX) will be empty anyhow.
    *ret\vt = #minScript_EMPTY
    *ret\q = 0
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following handles those standard run-time functions responsible for opening files.
Procedure.i minScript_RunTime_OPENFILE(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrString.STRING
  If numParamsPassed = 1
    If *params\index[0]\vt = #minScript_STRING
      *ret\vt = #minScript_INTEGER
      *ptrString = *params\index[0]
      Select fnID
        Case #minScript_STANDARDRUNTIME_CREATEFILE
          CompilerIf #PB_Compiler_OS = #PB_OS_Windows
            *ret\i = CreateFile(#PB_Any, *ptrString\s, #PB_File_SharedRead|#PB_File_SharedWrite)
          CompilerElse
            *ret\i = CreateFile(#PB_Any, *ptrString\s)
          CompilerEndIf
        Case #minScript_STANDARDRUNTIME_OPENFILE
          CompilerIf #PB_Compiler_OS = #PB_OS_Windows
            *ret\i = OpenFile(#PB_Any, *ptrString\s, #PB_File_SharedRead|#PB_File_SharedWrite)
          CompilerElse
            *ret\i = OpenFile(#PB_Any, *ptrString\s)
          CompilerEndIf
        Case #minScript_STANDARDRUNTIME_READFILE
          CompilerIf #PB_Compiler_OS = #PB_OS_Windows
            *ret\i = ReadFile(#PB_Any, *ptrString\s, #PB_File_SharedRead|#PB_File_SharedWrite)
          CompilerElse
            *ret\i = ReadFile(#PB_Any, *ptrString\s)
          CompilerEndIf
      EndSelect
      If *ret\i
        ;Need to add the fileID to the GCE.          
          CompilerIf #PB_Compiler_Thread
            LockMutex(minScript_RunTime_GCE_globals\threadMutex)
          CompilerEndIf
            If AddElement(minScript_RunTime_GCE_globals\fileIDs())
              minScript_RunTime_GCE_globals\fileIDs()\codeObject = codeObject
              minScript_RunTime_GCE_globals\fileIDs()\fileID = *ret\i
            Else ;Should free the memory just in case.
              CloseFile(*ret\i)
              *ret\i = 0
            EndIf
          CompilerIf #PB_Compiler_Thread
            UnlockMutex(minScript_RunTime_GCE_globals\threadMutex)
          CompilerEndIf
      EndIf
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_PEEK(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptr.st_minScript_memoryPointer
  If numParamsPassed = 1
    ;Change the parameter to an integer.
      result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
    If result = #minScript_OKAY
      *ptr = *params\index[0]\i
      If *ptr
        Select fnID
          Case #minScript_STANDARDRUNTIME_PEEKB
            *ret\vt = #minScript_INTEGER      
            *ret\i = *ptr\b
          Case #minScript_STANDARDRUNTIME_PEEKC
            *ret\vt = #minScript_INTEGER      
            *ret\i = *ptr\c
          Case #minScript_STANDARDRUNTIME_PEEKCURRENCY
            *ret\vt = #minScript_CURRENCY
            *ret\c = *ptr\q
          Case #minScript_STANDARDRUNTIME_PEEKF
            *ret\vt = #minScript_REAL
            *ret\r = *ptr\f
          Case #minScript_STANDARDRUNTIME_PEEKI
            *ret\vt = #minScript_INTEGER      
            *ret\i = *ptr\i
          Case #minScript_STANDARDRUNTIME_PEEKL
            *ret\vt = #minScript_INTEGER      
            *ret\i = *ptr\l
          Case #minScript_STANDARDRUNTIME_PEEKR
            *ret\vt = #minScript_REAL
            *ret\r = *ptr\d
          Case #minScript_STANDARDRUNTIME_PEEKS
            ;minScript_VarCopy() is the fastest way of copying a minScript variant string.
            result = minScript_VarCopy(*ret, *ptr, #minScript_VARDONOTFREEDESTINATIONVARIABLE)
          Case #minScript_STANDARDRUNTIME_PEEKW
            *ret\vt = #minScript_INTEGER      
            *ret\i = *ptr\w
        EndSelect
      Else
        result = #minScript_NULLPOINTER      
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_PEEKS(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptr.st_minScript_memoryPointer, nativeFormat, format
  CompilerIf Defined(PB_Compiler_Unicode,#PB_Constant)
    nativeFormat = #PB_Unicode
  CompilerElse
    nativeFormat = #PB_Ascii
  CompilerEndIf
  If numParamsPassed = 1 Or numParamsPassed = 2
    If numParamsPassed = 2
      ;Change the second parameter to an integer.
        minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
      Select *params\index[1]\i
        Case #PB_Ascii
          format = #PB_Ascii
        Case #PB_Unicode
          format = #PB_Unicode
        Case #PB_UTF8
          format = #PB_UTF8
        Default
          result = #minScript_INVALIDARGUMENT
      EndSelect
    Else
      format = nativeFormat
    EndIf
    If result = #minScript_OKAY
      ;Change the parameter to an integer.
        result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
      If result = #minScript_OKAY
        *ptr = *params\index[0]\i
        If *ptr
          If format = nativeFormat
            *ret\vt = #minScript_STRING
            *ret\sPtr = minScript_StringAlloc(*ptr)
            If *ret\sPtr = 0
              result = #minScript_OUTOFMEMORY
            EndIf
          Else
            result = minScript_VarSetString(*ret, PeekS(*ptr, -1, format), #False)
          EndIf
        Else
          result = #minScript_NULLPOINTER
        EndIf
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_POKE(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptr.st_minScript_memoryPointer, *ptrSource.st_minScript_memoryPointer
  If numParamsPassed = 2
    ;Change the first parameter to an integer.
      result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
    If result = #minScript_OKAY
      *ptr = *params\index[0]\i
      If *ptr
        Select fnID
          Case #minScript_STANDARDRUNTIME_POKEB
            result = minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
            If result = #minScript_OKAY
              *ptrSource = *params\index[1]
              *ptr\b = *ptrSource\i
            EndIf
          Case #minScript_STANDARDRUNTIME_POKEC
            result = minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
            If result = #minScript_OKAY
              *ptrSource = *params\index[1]
              *ptr\c = *ptrSource\i
            EndIf
          Case #minScript_STANDARDRUNTIME_POKECURRENCY
            result = minScript_VarChangeType(*params\index[1], #minScript_CURRENCY)
            If result = #minScript_OKAY
              *ptrSource = *params\index[1]
              *ptr\q = *ptrSource\q
            EndIf
          Case #minScript_STANDARDRUNTIME_POKEF
            result = minScript_VarChangeType(*params\index[1], #minScript_REAL)
            If result = #minScript_OKAY
              *ptrSource = *params\index[1]
              *ptr\f = *ptrSource\d
            EndIf
          Case #minScript_STANDARDRUNTIME_POKEL
            result = minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
            If result = #minScript_OKAY
              *ptrSource = *params\index[1]
              *ptr\l = *ptrSource\i
            EndIf
          Case #minScript_STANDARDRUNTIME_POKEI
            result = minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
            If result = #minScript_OKAY
              *ptrSource = *params\index[1]
              *ptr\i = *ptrSource\i
            EndIf
          Case #minScript_STANDARDRUNTIME_POKER
            result = minScript_VarChangeType(*params\index[1], #minScript_REAL)
            If result = #minScript_OKAY
              *ptrSource = *params\index[1]
              *ptr\d = *ptrSource\d
            EndIf
          Case #minScript_STANDARDRUNTIME_POKEW
            result = minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
            If result = #minScript_OKAY
              *ptrSource = *params\index[1]
              *ptr\w = *ptrSource\i
            EndIf
        EndSelect
      Else
        result = #minScript_NULLPOINTER      
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_POKES(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptr.st_minScript_memoryPointer, nativeFormat, format, *ptrString.STRING
  CompilerIf Defined(PB_Compiler_Unicode,#PB_Constant)
    nativeFormat = #PB_Unicode
  CompilerElse
    nativeFormat = #PB_Ascii
  CompilerEndIf
  If numParamsPassed = 2 Or numParamsPassed = 3
    If numParamsPassed = 3
      ;Change the third parameter to an integer.
        minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
      Select *params\index[2]\i
        Case #PB_Ascii
          format = #PB_Ascii
        Case #PB_Unicode
          format = #PB_Unicode
        Case #PB_UTF8
          format = #PB_UTF8
        Default
          result = #minScript_INVALIDARGUMENT
      EndSelect
    Else
      format = nativeFormat
    EndIf
    If result = #minScript_OKAY
      ;Change the parameters as appropriate.
        result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
      If result = #minScript_OKAY
        result = minScript_VarChangeType(*params\index[1], #minScript_STRING)
        If result = #minScript_OKAY
          If *params\index[0]\i
            *ret\vt = #minScript_INTEGER
            If format = nativeFormat
              ;We can use a straight memory copy here.
                *ptr = *params\index[1]\sPtr
                If *ptr
                  *ptr-SizeOf(LONG)
                  CopyMemory(*ptr+SizeOf(LONG), *params\index[0]\i, *ptr\l)
                  ;Add terminating null.
                    PokeC(*params\index[0]\i + *ptr\l, 0)
                  *ret\i = *ptr\l              
                EndIf
            Else
              *ptrString = *params\index[1]
              *ret\i = PokeS(*params\index[0]\i, *ptrString\s, -1, format)
            EndIf
          Else
            result = #minScript_NULLPOINTER
          EndIf
        EndIf
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Some of our file run-time functions which deal with reading strings and the like.
;A second optional parameter is the string format.
Procedure.i minScript_RunTime_READFILESTRINGS(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, format
  If numParamsPassed = 1 Or numParamsPassed = 2
    If *params\index[0]\vt = #minScript_INTEGER
      If numParamsPassed = 2
        ;Change the second parameter to an integer.
          minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
        Select *params\index[1]\i
          Case #PB_Ascii
            format = #PB_Ascii
          Case #PB_Unicode
            format = #PB_Unicode
          Case #PB_UTF8
            format = #PB_UTF8
          Default
            result = #minScript_INVALIDARGUMENT
        EndSelect
      Else
        CompilerIf Defined(PB_Compiler_Unicode, #PB_Constant)
          format = #PB_Unicode
        CompilerElse
          format = #PB_Ascii
        CompilerEndIf
      EndIf
      If result = #minScript_OKAY
        Select fnID
          Case #minScript_STANDARDRUNTIME_READCHARACTER
            *ret\vt = #minScript_INTEGER
            *ret\i = ReadCharacter(*params\index[0]\i, format)
          Case #minScript_STANDARDRUNTIME_READSTRING
            result = minScript_VarSetString(*ret, ReadString(*params\index[0]\i, format), #False)
        EndSelect
      EndIf
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_REALLOCATEMEMORY(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 2
    ;Change the first parameter to an integer.
      result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
    If result = #minScript_OKAY
      ;Change the second parameter to an integer.
        result = minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
      If result = #minScript_OKAY
        *ret\vt = #minScript_INTEGER
        *ret\i = ReAllocateMemory(*params\index[0]\i, *params\index[1]\i)
        If *ret\i
          ;We must alter any existing entry in the GCE or add a new entry as appropriate.
          CompilerIf #PB_Compiler_Thread
            LockMutex(minScript_RunTime_GCE_globals\threadMutex)
          CompilerEndIf
          If *params\index[0]\i ;Alter the existing GCE entry.
            ForEach minScript_RunTime_GCE_globals\memoryAllocations()
              If minScript_RunTime_GCE_globals\memoryAllocations()\address = *params\index[0]\i
                minScript_RunTime_GCE_globals\memoryAllocations()\address = *ret\i
                Break
              EndIf           
            Next
          Else ;Add a new GCE entry.
            If AddElement(minScript_RunTime_GCE_globals\memoryAllocations())
              minScript_RunTime_GCE_globals\memoryAllocations()\codeObject = codeObject
              minScript_RunTime_GCE_globals\memoryAllocations()\address = *ret\i
            Else ;Should free the memory just in case.
              FreeMemory(*ret\i)
              *ret\i = 0
            EndIf
          EndIf
          CompilerIf #PB_Compiler_Thread
            UnlockMutex(minScript_RunTime_GCE_globals\threadMutex)
          CompilerEndIf
        EndIf
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_STR(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, t1$
  If numParamsPassed = 1
    Select *params\index[0]\vt
      Case #minScript_INTEGER
        t1$ = Str(*params\index[0]\i)
      Case #minScript_CURRENCY
        t1$ = StrC(*params\index[0]\q)
      Case #minScript_REAL
        t1$ = StrD(*params\index[0]\r)
      Default
        ;Convert to an integer.
        result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
        If result = #minScript_OKAY
          t1$ = Str(*params\index[0]\i)
        EndIf
    EndSelect
    If result = #minScript_OKAY
      result = minScript_VarSetString(*ret, t1$, #False)
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_STRFIX(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, t1$
  If numParamsPassed = 1
    result = minScript_RunTime_STR(codeObject, fnID.l, numParamsPassed, *params, *ret)
  ElseIf numParamsPassed = 2
    ;Change the second parameter to an integer.
      minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
    If *params\index[1]\i < 0
      *params\index[1]\i = 0
    EndIf
    Select *params\index[0]\vt
      Case #minScript_INTEGER
        t1$ = StrD(*params\index[0]\i, *params\index[1]\i)
      Case #minScript_CURRENCY
        t1$ = StrC(*params\index[0]\q, *params\index[1]\i)
      Case #minScript_REAL
        t1$ = StrD(*params\index[0]\r, *params\index[1]\i)
      Default
        ;Convert to a real.
        result = minScript_VarChangeType(*params\index[0], #minScript_REAL)
        If result = #minScript_OKAY
          t1$ = StrD(*params\index[0]\r, *params\index[1]\i)
        EndIf
    EndSelect
    If result = #minScript_OKAY
      result = minScript_VarSetString(*ret, t1$, #False)
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following handles those standard run-time functions taking a single string parameter and returning an integer (e.g. "ASC").
Procedure.i minScript_RunTime_STRING2INTEGER(*code.st_minScriptCodeClass, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrChar.CHARACTER, *ptrLong.LONG, *ptrString.STRING
  If numParamsPassed = 1
    ;Change the parameter to a string.
      result = minScript_VarChangeType(*params\index[0], #minScript_STRING)
    If result = #minScript_OKAY
      *ret\vt = #minScript_INTEGER
      Select fnID
        Case #minScript_STANDARDRUNTIME_ASC
          *ptrChar = *params\index[0]\sPtr
          If *ptrChar
            *ret\i = *ptrChar\c
          EndIf
        Case #minScript_STANDARDRUNTIME_ISCLASS
          *ptrString = *params\index[0]
          If *ptrString\s
            *ret\i = minScriptClasses_HELPER_FindClass(*code, LCase(*ptrString\s))          
          EndIf
        Case #minScript_STANDARDRUNTIME_ISFUNCTION
          *ptrString = *params\index[0]
          If *ptrString\s
            *ret\i = minScriptCodeClass_HELPER_GetFunctionFromName(LCase(*ptrString\s), *code\currentScriptFunction\module)
          EndIf
        Case #minScript_STANDARDRUNTIME_ISRUNTIMEFUNCTION
          *ptrString = *params\index[0]
          If *ptrString\s
            If minScript_GetRuntimeFunctionFromName(*ptrString\s) = #minScript_OKAY
              *ret\i = #True
            Else
              *ret\i = #False
            EndIf
          EndIf
        Case #minScript_STANDARDRUNTIME_LEN
          *ptrLong = *params\index[0]\sPtr
          If *ptrLong
            *ptrLong-SizeOf(LONG)
            *ret\i = *ptrLong\l >> (SizeOf(CHARACTER)-1)
          EndIf
      EndSelect
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;The following handles those standard run-time functions taking a single string parameter and returning a string (e.g. "LCASE").
Procedure.i minScript_RunTime_STRING2STRING(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrString.STRING
  If numParamsPassed = 1
    ;Change the parameter to a string.
      result = minScript_VarChangeType(*params\index[0], #minScript_STRING)
    If result = #minScript_OKAY
      Select fnID
        Case #minScript_STANDARDRUNTIME_LCASE
          *ptrString = *params\index[0]
          result = minScript_VarSetString(*ret, LCase(*ptrString\s), #False)
        Case #minScript_STANDARDRUNTIME_UCASE
          *ptrString = *params\index[0]
          result = minScript_VarSetString(*ret, UCase(*ptrString\s), #False)
        Case #minScript_STANDARDRUNTIME_TOSTRING
          *ptrString = *params\index[0]
          result = minScript_VarSetString(*ret, *ptrString\s, #False)
        Case #minScript_STANDARDRUNTIME_TRIM
          *ptrString = *params\index[0]
          result = minScript_VarSetString(*ret, Trim(*ptrString\s), #False)
      EndSelect
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_STRINGBYTELENGTH(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptr.st_minScript_memoryPointer, nativeFormat, format, t1$, *ptrString.STRING
  CompilerIf Defined(PB_Compiler_Unicode,#PB_Constant)
    nativeFormat = #PB_Unicode
  CompilerElse
    nativeFormat = #PB_Ascii
  CompilerEndIf
  If numParamsPassed = 1 Or numParamsPassed <= 2
    If numParamsPassed = 2
      ;Change the second parameter to an integer.
        minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
      Select *params\index[1]\i
        Case #PB_Ascii
          format = #PB_Ascii
        Case #PB_Unicode
          format = #PB_Unicode
        Case #PB_UTF8
          format = #PB_UTF8
        Default
          result = #minScript_INVALIDARGUMENT
      EndSelect
    Else
      format = nativeFormat
    EndIf
    If result = #minScript_OKAY
      ;Change the parameter to a string.
        result = minScript_VarChangeType(*params\index[0], #minScript_STRING)
      If result = #minScript_OKAY
        *ret\vt = #minScript_INTEGER
        If format = nativeFormat
          ;In this case we can read the required number of bytes directly from the minScriptVariant string.
          *ptr = *params\index[0]\sPtr
          If *ptr
            *ptr-SizeOf(LONG)
            *ret\i = *ptr\l
          EndIf
        Else ;We use PB's StringByteLength() function.
          *ptrString = *params\index[0]
          *ret\i = StringByteLength(*ptrString\s, format)
        EndIf
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_STRINGFIELD(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrString1.STRING, *ptrString2.STRING
  If numParamsPassed = 3
    result = minScript_VarChangeType(*params\index[0], #minScript_STRING)
    If result = #minScript_OKAY
      *ptrString1 = *params\index[0]
      minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
      result = minScript_VarChangeType(*params\index[2], #minScript_STRING)
      If result = #minScript_OKAY
        *ptrString2 = *params\index[2]
        result = minScript_VarSetString(*ret, StringField(*ptrString1\s, *params\index[1]\i, *ptrString2\s), #False)
      EndIf
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_STRINGSET(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, *ptrString1.STRING, *ptrString2.STRING
  If numParamsPassed = 2 Or numParamsPassed = 3
    If *params\index[1]\vt = #minScript_INTEGER
      result = minScript_VarChangeType(*params\index[0], #minScript_STRING)
      If result = #minScript_OKAY
        *ptrString1 = *params\index[0]
        result = minScript_VarChangeType(*params\index[2], #minScript_STRING)
        If result = #minScript_OKAY
          *ptrString2 = *params\index[2]
          Select fnID
            Case #minScript_STANDARDRUNTIME_LSET
              result = minScript_VarSetString(*ret, LSet(*ptrString1\s, *params\index[1]\i, *ptrString2\s), #False)
            Case #minScript_STANDARDRUNTIME_RSET
              result = minScript_VarSetString(*ret, RSet(*ptrString1\s, *params\index[1]\i, *ptrString2\s), #False)
          EndSelect
        EndIf
      EndIf
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_TOCURRENCY(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 1
    ;Change the parameter to a currency.
      result = minScript_VarChangeType(*params\index[0], #minScript_CURRENCY)
    If result = #minScript_OKAY
      *ret\vt = #minScript_CURRENCY
      *ret\q = *params\index[0]\q
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_TOINTEGER(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 1
    ;Change the parameter to an integer.
      result = minScript_VarChangeType(*params\index[0], #minScript_INTEGER)
    If result = #minScript_OKAY
      *ret\vt = #minScript_INTEGER
      *ret\i = *params\index[0]\i
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_TOREAL(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 1
    ;Change the parameter to a real.
      result = minScript_VarChangeType(*params\index[0], #minScript_REAL)
    If result = #minScript_OKAY
      *ret\vt = #minScript_REAL
      *ret\r = *params\index[0]\r
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_TRUNCATE(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY
  If numParamsPassed = 1
    If *params\index[0]\vt = #minScript_CURRENCY
      *ret\vt = #minScript_INTEGER
      *ret\i = *params\index[0]\q/10000
    Else
      result = minScript_RunTime_TOINTEGER(codeObject, fnID.l, numParamsPassed, *params, *ret)
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_RunTime_TYPEOF(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, type$
  If numParamsPassed = 1
    Select fnID
      Case #minScript_STANDARDRUNTIME_TYPEOF
        *ret\vt = #minScript_INTEGER
        *ret\i = *params\index[0]\vt&#minScript_VARIANTMASK
      Case #minScript_STANDARDRUNTIME_TYPEOF2STRING
        type$ = minScriptCodeClass_DISASSEMBLY_VT2String(*params\index[0]\vt, codeObject)
        If type$ = ""
          type$ = "unknown"      
        EndIf
        result = minScript_VarSetString(*ret, type$, #False)
      Case #minScript_STANDARDRUNTIME_ISREFERENCETYPE
        *ret\vt = #minScript_INTEGER
        *ret\i = *params\index[0]\vt & #minScript_REFERENCE
    EndSelect
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;Some of our file run-time functions which deal with writing strings and the like.
;A third optional parameter is the string format.
Procedure.i minScript_RunTime_WRITEFILESTRINGS(codeObject.minScriptCode, fnID.l, numParamsPassed, *params.minScriptVariantParameters, *ret.minScriptVariant)
  Protected result = #minScript_OKAY, format, *ptrString.STRING
  If numParamsPassed = 2 Or numParamsPassed = 3
    If *params\index[0]\vt = #minScript_INTEGER
      If numParamsPassed = 3
        ;Change the second parameter to an integer.
          minScript_VarChangeType(*params\index[2], #minScript_INTEGER)
        Select *params\index[2]\i
          Case #PB_Ascii
            format = #PB_Ascii
          Case #PB_Unicode
            format = #PB_Unicode
          Case #PB_UTF8
            format = #PB_UTF8
          Default
            result = #minScript_INVALIDARGUMENT
        EndSelect
      Else
        CompilerIf Defined(PB_Compiler_Unicode, #PB_Constant)
          format = #PB_Unicode
        CompilerElse
          format = #PB_Ascii
        CompilerEndIf
      EndIf
      If result = #minScript_OKAY
        Select fnID
          Case #minScript_STANDARDRUNTIME_WRITECHARACTER
            minScript_VarChangeType(*params\index[1], #minScript_INTEGER)
            *ret\vt = #minScript_INTEGER
            *ret\i = WriteCharacter(*params\index[0]\i, *params\index[1]\i, format)
          Case #minScript_STANDARDRUNTIME_WRITESTRING
            result = minScript_VarChangeType(*params\index[1], #minScript_STRING)
            If result = #minScript_OKAY
              *ptrString = *params\index[1]
              *ret\vt = #minScript_INTEGER
              *ret\i = WriteString(*params\index[0]\i, *ptrString\s, format)
            EndIf
          Case #minScript_STANDARDRUNTIME_WRITESTRINGN
            result = minScript_VarChangeType(*params\index[1], #minScript_STRING)
            If result = #minScript_OKAY
              *ptrString = *params\index[1]
              *ret\vt = #minScript_INTEGER
              *ret\i = WriteStringN(*params\index[0]\i, *ptrString\s, format)
            EndIf
        EndSelect
      EndIf
    Else
      result = #minScript_INVALIDARGUMENT
    EndIf
  Else
    result = #minScript_INCORRECTNUMBEROFPARAMETERS
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-INTERNAL FUNCTIONS.

;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_InitStandardRuntime()
  Protected result = #minScript_OKAY
  ;Add standard runtime functions.
    result + minScript_AddRuntimeFunction("ALLOCATEMEMORY", #minScript_STANDARDRUNTIME_ALLOCATEMEMORY, @minScript_RunTime_INTEGER2INTEGER())
    result + minScript_AddRuntimeFunction("ASC", #minScript_STANDARDRUNTIME_ASC, @minScript_RunTime_STRING2INTEGER())
    result + minScript_AddRuntimeFunction("CHR", #minScript_STANDARDRUNTIME_CHR, @minScript_RunTime_INTEGER2STRING())
    result + minScript_AddRuntimeFunction("CLOSEFILE", #minScript_STANDARDRUNTIME_CLOSEFILE, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("CREATEFILE", #minScript_STANDARDRUNTIME_CREATEFILE, @minScript_RunTime_OPENFILE())
    result + minScript_AddRuntimeFunction("EOF", #minScript_STANDARDRUNTIME_EOF, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("FILESIZE", #minScript_STANDARDRUNTIME_FILESIZE, @minScript_RunTime_FILE2())
    result + minScript_AddRuntimeFunction("FINDSTRING", #minScript_STANDARDRUNTIME_FINDSTRING, @minScript_RunTime_FINDSTRING())
    result + minScript_AddRuntimeFunction("FREEMEMORY", #minScript_STANDARDRUNTIME_FREEMEMORY, @minScript_RunTime_INTEGER2INTEGER())
    result + minScript_AddRuntimeFunction("GETEXTENSIONPART", #minScript_STANDARDRUNTIME_GETEXTENSIONPART, @minScript_RunTime_FILE2())
    result + minScript_AddRuntimeFunction("GETFILEPART", #minScript_STANDARDRUNTIME_GETFILEPART, @minScript_RunTime_FILE2())
    result + minScript_AddRuntimeFunction("GETPATHPART", #minScript_STANDARDRUNTIME_GETPATHPART, @minScript_RunTime_FILE2())
    result + minScript_AddRuntimeFunction("HEX", #minScript_STANDARDRUNTIME_HEX, @minScript_RunTime_INTEGER2STRING())
    result + minScript_AddRuntimeFunction("ISCLASS", #minScript_STANDARDRUNTIME_ISCLASS, @minScript_RunTime_STRING2INTEGER())
    result + minScript_AddRuntimeFunction("ISFUNCTION", #minScript_STANDARDRUNTIME_ISFUNCTION, @minScript_RunTime_STRING2INTEGER())
    result + minScript_AddRuntimeFunction("ISREFERENCETYPE", #minScript_STANDARDRUNTIME_ISREFERENCETYPE, @minScript_RunTime_TYPEOF())
    result + minScript_AddRuntimeFunction("ISRUNTIMEFUNCTION", #minScript_STANDARDRUNTIME_ISRUNTIMEFUNCTION, @minScript_RunTime_STRING2INTEGER())
    result + minScript_AddRuntimeFunction("LCASE", #minScript_STANDARDRUNTIME_LCASE, @minScript_RunTime_STRING2STRING())
    result + minScript_AddRuntimeFunction("LEFT", #minScript_STANDARDRUNTIME_LEFT, @minScript_RunTime_LEFTRIGHT())
    result + minScript_AddRuntimeFunction("LEN", #minScript_STANDARDRUNTIME_LEN, @minScript_RunTime_STRING2INTEGER())
    result + minScript_AddRuntimeFunction("LOC", #minScript_STANDARDRUNTIME_LOC, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("LOF", #minScript_STANDARDRUNTIME_LOF, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("LSET", #minScript_STANDARDRUNTIME_LSET, @minScript_RunTime_STRINGSET())
    result + minScript_AddRuntimeFunction("MID", #minScript_STANDARDRUNTIME_MID, @minScript_RunTime_MID())
    result + minScript_AddRuntimeFunction("NOTHING", #minScript_STANDARDRUNTIME_NOTHING, @minScript_RunTime_NOTHING())
    result + minScript_AddRuntimeFunction("OPENFILE", #minScript_STANDARDRUNTIME_OPENFILE, @minScript_RunTime_OPENFILE())
    result + minScript_AddRuntimeFunction("PEEKB", #minScript_STANDARDRUNTIME_PEEKB, @minScript_RunTime_PEEK())
    result + minScript_AddRuntimeFunction("PEEKC", #minScript_STANDARDRUNTIME_PEEKC, @minScript_RunTime_PEEK())
    result + minScript_AddRuntimeFunction("PEEKCURRENCY", #minScript_STANDARDRUNTIME_PEEKCURRENCY, @minScript_RunTime_PEEK())
    result + minScript_AddRuntimeFunction("PEEKF", #minScript_STANDARDRUNTIME_PEEKF, @minScript_RunTime_PEEK())
    result + minScript_AddRuntimeFunction("PEEKI", #minScript_STANDARDRUNTIME_PEEKI, @minScript_RunTime_PEEK())
    result + minScript_AddRuntimeFunction("PEEKL", #minScript_STANDARDRUNTIME_PEEKL, @minScript_RunTime_PEEK())
    result + minScript_AddRuntimeFunction("PEEKR", #minScript_STANDARDRUNTIME_PEEKR, @minScript_RunTime_PEEK())
    result + minScript_AddRuntimeFunction("PEEKS", #minScript_STANDARDRUNTIME_PEEKS, @minScript_RunTime_PEEKS())
    result + minScript_AddRuntimeFunction("PEEKW", #minScript_STANDARDRUNTIME_PEEKW, @minScript_RunTime_PEEK())
    result + minScript_AddRuntimeFunction("POKEB", #minScript_STANDARDRUNTIME_POKEB, @minScript_RunTime_POKE())
    result + minScript_AddRuntimeFunction("POKEC", #minScript_STANDARDRUNTIME_POKEC, @minScript_RunTime_POKE())
    result + minScript_AddRuntimeFunction("POKECURRENCY", #minScript_STANDARDRUNTIME_POKECURRENCY, @minScript_RunTime_POKE())
    result + minScript_AddRuntimeFunction("POKEF", #minScript_STANDARDRUNTIME_POKEF, @minScript_RunTime_POKE())
    result + minScript_AddRuntimeFunction("POKEI", #minScript_STANDARDRUNTIME_POKEI, @minScript_RunTime_POKE())
    result + minScript_AddRuntimeFunction("POKEL", #minScript_STANDARDRUNTIME_POKEL, @minScript_RunTime_POKE())
    result + minScript_AddRuntimeFunction("POKER", #minScript_STANDARDRUNTIME_POKER, @minScript_RunTime_POKE())
    result + minScript_AddRuntimeFunction("POKES", #minScript_STANDARDRUNTIME_POKES, @minScript_RunTime_POKES())
    result + minScript_AddRuntimeFunction("POKEW", #minScript_STANDARDRUNTIME_POKEW, @minScript_RunTime_POKE())
    result + minScript_AddRuntimeFunction("READBYTE", #minScript_STANDARDRUNTIME_READBYTE, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("READCHARACTER", #minScript_STANDARDRUNTIME_READCHARACTER, @minScript_RunTime_READFILESTRINGS())
    result + minScript_AddRuntimeFunction("READDATA", #minScript_STANDARDRUNTIME_READDATA, @minScript_RunTime_FILEDATA())
    result + minScript_AddRuntimeFunction("READDOUBLE", #minScript_STANDARDRUNTIME_READDOUBLE, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("READFILE", #minScript_STANDARDRUNTIME_READFILE, @minScript_RunTime_OPENFILE())
    result + minScript_AddRuntimeFunction("READFLOAT", #minScript_STANDARDRUNTIME_READFLOAT, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("READINTEGER", #minScript_STANDARDRUNTIME_READINTEGER, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("READLONG", #minScript_STANDARDRUNTIME_READLONG, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("READSTRING", #minScript_STANDARDRUNTIME_READSTRING, @minScript_RunTime_READFILESTRINGS())
    result + minScript_AddRuntimeFunction("READSTRINGFORMAT", #minScript_STANDARDRUNTIME_READSTRINGFORMAT, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("READWORD", #minScript_STANDARDRUNTIME_READWORD, @minScript_RunTime_FILE1())
    result + minScript_AddRuntimeFunction("REALLOCATEMEMORY", #minScript_STANDARDRUNTIME_REALLOCATEMEMORY, @minScript_RunTime_REALLOCATEMEMORY())
    result + minScript_AddRuntimeFunction("RIGHT", #minScript_STANDARDRUNTIME_RIGHT, @minScript_RunTime_LEFTRIGHT())
    result + minScript_AddRuntimeFunction("RSET", #minScript_STANDARDRUNTIME_RSET, @minScript_RunTime_STRINGSET())
    result + minScript_AddRuntimeFunction("SIZEOF", #minScript_STANDARDRUNTIME_SIZEOF, @minScript_RunTime_INTEGER2INTEGER())
    result + minScript_AddRuntimeFunction("SPACE", #minScript_STANDARDRUNTIME_SPACE, @minScript_RunTime_INTEGER2STRING())
    result + minScript_AddRuntimeFunction("STR", #minScript_STANDARDRUNTIME_STR, @minScript_RunTime_STR())
    result + minScript_AddRuntimeFunction("STRFIX", #minScript_STANDARDRUNTIME_STRFIX, @minScript_RunTime_STRFIX())
    result + minScript_AddRuntimeFunction("STRINGBYTELENGTH", #minScript_STANDARDRUNTIME_STRINGBYTELENGTH, @minScript_RunTime_STRINGBYTELENGTH())
    result + minScript_AddRuntimeFunction("STRINGFIELD", #minScript_STANDARDRUNTIME_STRINGFIELD, @minScript_RunTime_STRINGFIELD())
    result + minScript_AddRuntimeFunction("THROWERROR", #minScript_STANDARDRUNTIME_THROWERROR, @minScript_RunTime_INTEGER2INTEGER())
    result + minScript_AddRuntimeFunction("TOCURRENCY", #minScript_STANDARDRUNTIME_TOCURRENCY, @minScript_RunTime_TOCURRENCY())
    result + minScript_AddRuntimeFunction("TOINTEGER", #minScript_STANDARDRUNTIME_TOINTEGER, @minScript_RunTime_TOINTEGER())
    result + minScript_AddRuntimeFunction("TOREAL", #minScript_STANDARDRUNTIME_TOREAL, @minScript_RunTime_TOREAL())
    result + minScript_AddRuntimeFunction("TOSTRING", #minScript_STANDARDRUNTIME_TOSTRING, @minScript_RunTime_STRING2STRING())
    result + minScript_AddRuntimeFunction("TRIM", #minScript_STANDARDRUNTIME_TRIM, @minScript_RunTime_STRING2STRING())
    result + minScript_AddRuntimeFunction("TRUNCATE", #minScript_STANDARDRUNTIME_TRUNCATE, @minScript_RunTime_TRUNCATE())
    result + minScript_AddRuntimeFunction("TYPEOF", #minScript_STANDARDRUNTIME_TYPEOF, @minScript_RunTime_TYPEOF())
    result + minScript_AddRuntimeFunction("TYPEOF2STRING", #minScript_STANDARDRUNTIME_TYPEOF2STRING, @minScript_RunTime_TYPEOF())
    result + minScript_AddRuntimeFunction("UCASE", #minScript_STANDARDRUNTIME_UCASE, @minScript_RunTime_STRING2STRING())
    result + minScript_AddRuntimeFunction("WRITEBYTE", #minScript_STANDARDRUNTIME_WRITEBYTE, @minScript_RunTime_FILE3())
    result + minScript_AddRuntimeFunction("WRITECHARACTER", #minScript_STANDARDRUNTIME_WRITECHARACTER, @minScript_RunTime_WRITEFILESTRINGS())
    result + minScript_AddRuntimeFunction("WRITEDOUBLE", #minScript_STANDARDRUNTIME_WRITEDOUBLE, @minScript_RunTime_FILE3())
    result + minScript_AddRuntimeFunction("WRITEFLOAT", #minScript_STANDARDRUNTIME_WRITEFLOAT, @minScript_RunTime_FILE3())
    result + minScript_AddRuntimeFunction("WRITEINTEGER", #minScript_STANDARDRUNTIME_WRITEINTEGER, @minScript_RunTime_FILE3())
    result + minScript_AddRuntimeFunction("WRITELONG", #minScript_STANDARDRUNTIME_WRITELONG, @minScript_RunTime_FILE3())
    result + minScript_AddRuntimeFunction("WRITESTRING", #minScript_STANDARDRUNTIME_WRITESTRING, @minScript_RunTime_WRITEFILESTRINGS())
    result + minScript_AddRuntimeFunction("WRITESTRINGFORMAT", #minScript_STANDARDRUNTIME_WRITESTRINGFORMAT, @minScript_RunTime_FILE3())
    result + minScript_AddRuntimeFunction("WRITESTRINGN", #minScript_STANDARDRUNTIME_WRITESTRINGN, @minScript_RunTime_WRITEFILESTRINGS())
    result + minScript_AddRuntimeFunction("WRITEWORD", #minScript_STANDARDRUNTIME_WRITEWORD, @minScript_RunTime_FILE3())
  If result = #minScript_OKAY
    ;Create a mutex to protect access to the GCE data if appropriate.
      CompilerIf #PB_Compiler_Thread
        minScript_RunTime_GCE_globals\threadMutex = CreateMutex()
        If minScript_RunTime_GCE_globals\threadMutex = 0
          result = #minScript_OUTOFMEMORY
        EndIf
      CompilerEndIf
    If result = #minScript_OKAY
      ;Add appropriate GC extensions.
        minScript_AddGCExtension(@minScript_RunTime_GCE())
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;-GC Extensions required by standard run-time.

;/////////////////////////////////////////////////////////////////////////////////
;The following GC extension covers the freeing of memory allocated by our ALLOCATEMEMORY() function which may not have been disposed of through script.
Procedure minScript_RunTime_GCE(codeObject, uMsg)
  Select uMsg
    Case #minScript_GCEXTENSION_CODEOBJECTHASBEENCREATED
      ;Not needed.
    Default ; #minScript_GCEXTENSION_CODEOBJECTHASBEENRESET or #minScript_GCEXTENSION_CODEOBJECTHASBEENDELETED
      ;We simply loop through all outstanding resources and free those belonging to the given code object.
      CompilerIf #PB_Compiler_Thread
        LockMutex(minScript_RunTime_GCE_globals\threadMutex)
      CompilerEndIf
        ForEach minScript_RunTime_GCE_globals\memoryAllocations()
          If minScript_RunTime_GCE_globals\memoryAllocations()\codeObject = codeObject
            FreeMemory(minScript_RunTime_GCE_globals\memoryAllocations()\address)
            DeleteElement(minScript_RunTime_GCE_globals\memoryAllocations())
          EndIf           
        Next
        ForEach minScript_RunTime_GCE_globals\FileIDs()
          If minScript_RunTime_GCE_globals\fileIDs()\codeObject = codeObject
            CloseFile(minScript_RunTime_GCE_globals\fileIDs()\fileID)
            DeleteElement(minScript_RunTime_GCE_globals\fileIDs())
          EndIf           
        Next
      CompilerIf #PB_Compiler_Thread
        UnlockMutex(minScript_RunTime_GCE_globals\threadMutex)
      CompilerEndIf
  EndSelect
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////

; IDE Options = PureBasic 5.62 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 122
; FirstLine = 122
; Folding = -------
; EnableThread
; Executable = d.exe.dll.exe.dll
; EnableUnicode