﻿;/////////////////////////////////////////////////////////////////////////////////
;minScript.
;==========
;
;Variant utilities Module.
;
;© Big Eared Software 2017.
;==========================
;/////////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////////
;NOTES.
;
;This module contains utility functions of use by the execution engine and by the minScript compiler.
;/////////////////////////////////////////////////////////////////////////////////


;-INCLUDES.

;-DECLARES.
  Declare.i minScript_VarUtils_PromoteVariablesToNumeric(*op1.minScriptVariant, *op2.minScriptVariant)



;/////////////////////////////////////////////////////////////////////////////////
;The following function performs a binary arithmetic operation on the given minScript variants (leaving the result in *op1).
;The operation is dictated by the value of opCode which assumes one of the #MS_OP_ADD values etc.
;The 2 variants will be promoted as appropriate.
Procedure.i minScript_VarUtils_ArithmeticOP(*op1.minScriptVariant, *op2.minScriptVariant, opCode)
  Protected result = #minScript_OKAY, varType
  ;First job is to promote the variables to numerics.
    varType = minScript_VarUtils_PromoteVariablesToNumeric(*op1, *op2)
  Select varType
    Case #minScript_INTEGER
      Select opCode
        Case #MS_OP_ADD
          *op1\i + *op2\i
        Case #MS_OP_SUB
          *op1\i - *op2\i
        Case #MS_OP_MUL
          *op1\i * *op2\i
        Case #MS_OP_DIV
          If *op2\i <> 0
            *op1\i / *op2\i
          Else
            result = #minScript_DIVISIONBYZERO
          EndIf
      EndSelect
    Case #minScript_CURRENCY
      Select opCode
        Case #MS_OP_ADD
          *op1\q + *op2\q
        Case #MS_OP_SUB
          *op1\q - *op2\q
        Case #MS_OP_MUL
          *op1\q = MultC(*op1\q, *op2\q)
        Case #MS_OP_DIV
          If *op2\q <> 0
            *op1\q = DivC(*op1\q, *op2\q)
          Else
            result = #minScript_DIVISIONBYZERO
          EndIf
      EndSelect
    Case #minScript_REAL
      Select opCode
        Case #MS_OP_ADD
          *op1\r + *op2\r
        Case #MS_OP_SUB
          *op1\r - *op2\r
        Case #MS_OP_MUL
          *op1\r * *op2\r
        Case #MS_OP_DIV
          If *op2\r <> 0
            *op1\r / *op2\r
          Else
            result = #minScript_DIVISIONBYZERO
          EndIf
      EndSelect
    Default
      result = varType ;An error code.
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function initialises the given minScript variant with an appropriate numeric type.
;We pass a pointer to a string containing what should be a number. 
;*var will be initialised with either integer type, real type or currency type.
Procedure.i minScript_VarUtils_InitNumericVariant(*var.minScriptVariant, *ptrValue.CHARACTER)
  Protected result = #minScript_OKAY, dblValue.d, blnIsCurrency, *ptrString.STRING
  If *ptrValue
    ;Determine if we have an integer.
      If minScript_SymbolsIsInteger(*ptrValue, *var) = #minScript_OKAY
        *var\vt = #minScript_INTEGER
      ;Otherwise investigate a currency/real.
      ElseIf minScript_SymbolsIsReal(*ptrValue, @dblValue, @blnIsCurrency) = #minScript_OKAY
        If blnIsCurrency
          *var\vt = #minScript_CURRENCY
          *ptrString = @*ptrValue
          *var\c = ValC(*ptrString\s)
        Else ;Real.
          *var\vt = #minScript_REAL
          *var\r = dblValue
        EndIf
      Else
        result = #minScript_INVALIDDATATYPE
      EndIf
  Else
    result = #minScript_INVALIDPARAMETER
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function returns #True if the given variant holds a value not equating to 'Null'.
;Note that an empty variant will be recorded as being null.
Procedure.i minScript_VarUtils_IsNotNull(*var.minScriptVariant)
  Protected blnNotNullRecorded
  ;Do we have a reference type.
  If *var\vt&#minScript_REFERENCE 
    If *var\refPtr
      blnNotNullRecorded = #True
    EndIf
  Else
    Select *var\vt
      Case #minScript_INTEGER
        If *var\i
          blnNotNullRecorded = #True
        EndIf
      Case #minScript_CURRENCY
        If *var\q
          blnNotNullRecorded = #True
        EndIf
      Case #minScript_REAL
        If *var\r
          blnNotNullRecorded = #True
        EndIf
      Case #minScript_STRING
        If *var\sPtr And PeekS(*var\sPtr) <> ""
          blnNotNullRecorded = #True
        EndIf
    EndSelect
  EndIf
  ProcedureReturn blnNotNullRecorded
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The operation is dictated by the value of opCode which assumes one of the #MS_OP_AND values etc. (Leaving the result in *op1)
Procedure.i minScript_VarUtils_LogicalOP(*op1.minScriptVariant, *op2.minScriptVariant, opCode)
  Protected result = #minScript_OKAY, blnIsNotNull1, blnIsNotNull2, tResult
  blnIsNotNull1 = minScript_VarUtils_IsNotNull(*op1)
  blnIsNotNull2 = minScript_VarUtils_IsNotNull(*op2)
  ;Clear the 2 variants and proceed only if there is no error.
    result = minScript_VarClear(*op1)
    If result = #minScript_OKAY
      result = minScript_VarClear(*op2)
      If result = #minScript_OKAY
        *op1\vt = #minScript_INTEGER
        Select opCode
          Case #MS_OP_AND
            *op1\i = blnIsNotNull1 & blnIsNotNull2
          Case #MS_OP_OR
            *op1\i = blnIsNotNull1 | blnIsNotNull2
          Case #MS_OP_XOR
            *op1\i = blnIsNotNull1 ! blnIsNotNull2
        EndSelect
      EndIf
    EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
Procedure.i minScript_VarUtils_Negate(*var.minScriptVariant)
  Protected result = #minScript_OKAY, numVar.minScriptVariant
  If *var\vt&#minScript_REFERENCE = 0
    Select *var\vt
      Case #minScript_INTEGER
        *var\i = -*var\i
      Case #minScript_CURRENCY
        *var\q = -*var\q
      Case #minScript_REAL
        *var\r = -*var\r
      Case #minScript_STRING
        ;In this case we attempt to convert to a numeric type and then negate appropriately.
        If *var\sPtr
          result = minScript_VarUtils_InitNumericVariant(numVar, *var\sPtr)
          If result = #minScript_OKAY
            minScript_VarUtils_Negate(numVar)
            minScript_StringFree(*var\sPtr)
            *var\vt = numVar\vt
            *var\q = numVar\q
          EndIf
        Else
          result = #minScript_INVALIDDATATYPE
        EndIf
      Default
        result = #minScript_INVALIDDATATYPE
    EndSelect
  Else
    result = #minScript_INVALIDDATATYPE
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function performs a numeric comparison on the given minScript variants (leaving the result in *op1).
;The operation is dictated by the value of opCode which assumes one of the #MS_OP_EQU values etc.
;The 2 variants will be promoted as appropriate.
Procedure.i minScript_VarUtils_NumericComparison(*op1.minScriptVariant, *op2.minScriptVariant, opCode)
  Protected result = #minScript_OKAY, varType
  ;First job is to promote the variables to numerics.
    varType = minScript_VarUtils_PromoteVariablesToNumeric(*op1, *op2)
  *op1\vt = #minScript_INTEGER
  Select varType
    Case #minScript_INTEGER
      Select opCode
        Case #MS_OP_EQU
          If *op1\i = *op2\i
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_LT
          If *op1\i < *op2\i
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_LTE
          If *op1\i <= *op2\i
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_GT
          If *op1\i > *op2\i
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_GTE
          If *op1\i >= *op2\i
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_NEQ
          If *op1\i <> *op2\i
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
      EndSelect
    Case #minScript_CURRENCY
      Select opCode
        Case #MS_OP_EQU
          If *op1\q = *op2\q
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_LT
          If *op1\q < *op2\q
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_LTE
          If *op1\q <= *op2\q
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_GT
          If *op1\q > *op2\q
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_GTE
          If *op1\q >= *op2\q
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_NEQ
          If *op1\q <> *op2\q
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
      EndSelect
    Case #minScript_REAL
      Select opCode
        Case #MS_OP_EQU
          If *op1\r = *op2\r
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_LT
          If *op1\r < *op2\r
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_LTE
          If *op1\r <= *op2\r
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_GT
          If *op1\r > *op2\r
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_GTE
          If *op1\r >= *op2\r
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
        Case #MS_OP_NEQ
          If *op1\r <> *op2\r
            *op1\i = #True
          Else
            *op1\i = #False
          EndIf
      EndSelect
    Default
      result = varType ;An error code.
  EndSelect
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////


;/////////////////////////////////////////////////////////////////////////////////
;The following function examines the two variants provided prior to some arithmetic operation and, if appropriate, promotes
;one to match the other's type.
;Returns either a minScript variable vt type : #minScript_INTEGER, #minScript_CURRENCY, #minScript_REAL or an error code (negative).
Procedure.i minScript_VarUtils_PromoteVariablesToNumeric(*op1.minScriptVariant, *op2.minScriptVariant)
  Protected result = #minScript_OKAY, var.minScriptVariant
  ;Convert any reference types to integers or strings to appropriate numerics.
    If *op1\vt&#minScript_REFERENCE 
      ;We must call the relevant \Clear() method.
        If *op1\refPtr
          result = *op1\refPtr\Clear()
        EndIf
      *op1\vt = #minScript_INTEGER
    ElseIf *op1\vt = #minScript_STRING
      If *op1\sPtr
        result = minScript_VarUtils_InitNumericVariant(var, *op1\sPtr)
        If result = #minScript_OKAY
          minScript_StringFree(*op1\sPtr)
          *op1\vt = var\vt
          *op1\q = var\q
        EndIf
      Else
        result = #minScript_INVALIDDATATYPE
      EndIf
    EndIf
    If result = #minScript_OKAY
      If *op2\vt&#minScript_REFERENCE 
        ;We must call the relevant \Clear() method.
          If *op2\refPtr
            result = *op2\refPtr\Clear()
          EndIf
        *op2\vt = #minScript_INTEGER
      ElseIf *op2\vt = #minScript_STRING
        If *op2\sPtr
          result = minScript_VarUtils_InitNumericVariant(var, *op2\sPtr)
          If result = #minScript_OKAY
            minScript_StringFree(*op2\sPtr)
            *op2\vt = var\vt
            *op2\q = var\q
          EndIf
        Else
          result = #minScript_INVALIDDATATYPE
        EndIf
      EndIf
    EndIf
  If result = #minScript_OKAY
    If *op1\vt = *op2\vt ;We can avoid any promotion here.
      Select *op1\vt
        Case #minScript_EMPTY ;If just one operand is empty then the else clause below will promote as appropriate.
          ;We simply convert both to zero integers.
            *op1\vt = #minScript_INTEGER
            *op2\vt = #minScript_INTEGER
        Case #minScript_INTEGER, #minScript_CURRENCY, #minScript_REAL
          result = *op1\vt    
        Default 
          result = #minScript_INCOMPATIBLEDATATYPES
      EndSelect
    Else
      If *op1\vt = #minScript_REAL
        minScript_VarChangeType(*op2, #minScript_REAL)
      ElseIf *op2\vt = #minScript_REAL
        minScript_VarChangeType(*op1, #minScript_REAL)
      ElseIf *op1\vt = #minScript_CURRENCY
        minScript_VarChangeType(*op2, #minScript_CURRENCY)
      ElseIf *op2\vt = #minScript_CURRENCY
        minScript_VarChangeType(*op1, #minScript_CURRENCY)
      Else ;One operand might be empty.
        minScript_VarChangeType(*op1, #minScript_INTEGER)
        minScript_VarChangeType(*op2, #minScript_INTEGER)
      EndIf ;Only option remaining is that both operands are integer type and that is covered above.
      result = *op1\vt
    EndIf
  EndIf
  ProcedureReturn result
EndProcedure
;/////////////////////////////////////////////////////////////////////////////////



; IDE Options = PureBasic 5.62 (Windows - x64)
; ExecutableFormat = Shared dll
; CursorPosition = 415
; FirstLine = 395
; Folding = --
; Executable = d.exe.dll.exe.dll
; CompileSourceDirectory